"""Create plots of a dataset."""
import itertools
from pathlib import Path
from typing import Any, Callable, List, Optional, Tuple, Union
from typing_extensions import Literal

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from matplotlib import figure, legend
from sklearn.manifold import TSNE  # type: ignore[attr-defined]

from ethicml.metrics.metric import Metric
from ethicml.utility import DataTuple, Results

from .common import DataEntry, LegendType, PlotDef, PlotType, errorbox, scatter

__all__ = [
    "plot_results",
    "single_plot",
    "save_2d_plot",
    "save_label_plot",
    "save_jointplot",
    "save_multijointplot",
]

MARKERS = ["s", "p", "P", "*", "+", "x", "o", "v"]


def maybe_tsne(data: DataTuple) -> Tuple[pd.DataFrame, str, str]:
    columns = data.x.columns

    if len(columns) > 2:
        tsne_embeddings = TSNE(n_components=2, random_state=0).fit_transform(data.x)
        amalgamated = pd.concat(
            [pd.DataFrame(tsne_embeddings, columns=["tsne1", "tsne2"]), data.s, data.y],
            axis="columns",
        )
        x1_name = "tsne1"
        x2_name = "tsne2"
    else:
        amalgamated = data.data
        x1_name = f"{columns[0]}"
        x2_name = f"{columns[1]}"
    return amalgamated, x1_name, x2_name


def save_2d_plot(data: DataTuple, filepath: str) -> None:
    """Make 2D plot."""
    file_path = Path(filepath)

    amalgamated, x1_name, x2_name = maybe_tsne(data)

    plot = sns.scatterplot(  # type: ignore[attr-defined]
        x=x1_name,
        y=x2_name,
        hue=data.y.name,
        palette="Set2",
        data=amalgamated,
        style=data.s.name,
        legend="full",
    )

    file_path.parent.mkdir(exist_ok=True)
    plot.figure.savefig(file_path)
    plt.clf()


def save_jointplot(data: DataTuple, filepath: str, dims: Tuple[int, int] = (0, 1)) -> None:
    """Make joint plot."""
    file_path = Path(filepath)
    columns = data.x.columns

    amalgamated = pd.concat([data.x, data.y], axis="columns")

    plot = sns.jointplot(x=columns[dims[0]], y=columns[dims[1]], data=amalgamated, kind="kde")  # type: ignore[attr-defined]

    file_path.parent.mkdir(exist_ok=True)
    plot.savefig(file_path)
    plt.clf()


def multivariate_grid(
    col_x: str,
    col_y: str,
    sens_col: str,
    outcome_col: str,
    df: pd.DataFrame,
    scatter_alpha: float = 0.5,
) -> None:
    def colored_scatter(x: Any, y: Any, c: Optional[str] = None) -> Callable[[Any], None]:
        def scatter(*args: Any, **kwargs: Any) -> None:
            args = (x, y)
            if c is not None:
                kwargs['c'] = c
            kwargs['alpha'] = scatter_alpha
            plt.scatter(*args, **kwargs)

        return scatter

    sns.set_palette("husl")  # type: ignore[attr-defined]

    g = sns.JointGrid(x=col_x, y=col_y, data=df)  # type: ignore[attr-defined]
    color = None
    legends = []
    for name, df_group in df.groupby([sens_col, outcome_col]):
        legends.append(f"S={name[0]}, Y={name[1]}")
        g.plot_joint(colored_scatter(df_group[col_x], df_group[col_y], color))
        sns.distplot(df_group[col_x].values, ax=g.ax_marg_x, color=color)  # type: ignore[attr-defined]
        sns.distplot(df_group[col_y].values, ax=g.ax_marg_y, vertical=True)  # type: ignore[attr-defined]
    # Do also global Hist:
    # sns.distplot(df[col_x].values, ax=g.ax_marg_x, color='grey')
    # sns.distplot(df[col_y].values.ravel(), ax=g.ax_marg_y, color='grey', vertical=True)
    plt.legend(legends)


def save_multijointplot(data: DataTuple, filepath: str) -> None:
    """Make joint plot."""
    file_path = Path(filepath)
    data.x.columns

    amalgamated, x1_name, x2_name = maybe_tsne(data)

    multivariate_grid(
        col_x=x1_name,
        col_y=x2_name,
        sens_col=str(data.s.name),
        outcome_col=str(data.y.name),
        df=amalgamated,
    )

    file_path.parent.mkdir(exist_ok=True)
    plt.savefig(file_path)
    plt.clf()


def make_gif(files: List[str], name: str = "movie") -> None:
    """Make GIF."""
    import imageio

    images = [imageio.imread(filename) for filename in files]
    imageio.mimsave(f"{name}.gif", images)


def save_label_plot(data: DataTuple, filename: str) -> None:
    """Make label plot."""
    file_path = Path(filename)

    # Only consider 1 sens attr for now
    s_values = data.s.value_counts() / data.s.count()

    s_0_val = s_values[0]
    s_1_val = s_values[1]

    s_0_label = s_values.index.min()
    s_1_label = s_values.index.max()

    y_s0 = data.y[data.s == 0].value_counts() / data.y[data.s == 0].count()
    y_s1 = data.y[data.s == 1].value_counts() / data.y[data.s == 1].count()

    y_0_label = y_s0.index[0]
    y_1_label = y_s0.index[1]

    plt.style.use("seaborn-pastel")
    # plt.xkcd()

    fig, plot = plt.subplots()

    quadrant1 = plot.bar(
        0,
        height=y_s0[y_0_label] * 100,  # type: ignore[call-overload]
        width=s_0_val * 100,
        align="edge",
        edgecolor="black",
        color="C0",
    )
    quadrant2 = plot.bar(
        s_0_val * 100,
        height=y_s1[y_0_label] * 100,  # type: ignore[call-overload]
        width=s_1_val * 100,
        align="edge",
        edgecolor="black",
        color="C1",
    )
    quadrant3 = plot.bar(
        0,
        height=y_s0[y_1_label] * 100,  # type: ignore[call-overload]
        width=s_0_val * 100,
        bottom=y_s0[y_0_label] * 100,  # type: ignore[call-overload]
        align="edge",
        edgecolor="black",
        color="C2",
    )
    quadrant4 = plot.bar(
        s_0_val * 100,
        height=y_s1[y_1_label] * 100,  # type: ignore[call-overload]
        width=s_1_val * 100,
        bottom=y_s1[y_0_label] * 100,  # type: ignore[call-overload]
        align="edge",
        edgecolor="black",
        color="C3",
    )

    plot.set_ylim(0, 100)
    plot.set_xlim(0, 100)
    plot.set_ylabel(f"Percent {data.y.name}=y")
    plot.set_xlabel(f"Percent {data.s.name}=s")
    plot.set_title("Dataset Composition by class and sensitive attribute")

    plot.legend(
        [quadrant1, quadrant2, quadrant3, quadrant4],
        [
            f"y={y_0_label}, s={s_0_label}",
            f"y={y_0_label}, s={s_1_label}",
            f"y={y_1_label}, s={s_0_label}",
            f"y={y_1_label}, s={s_1_label}",
        ],
    )

    file_path.parent.mkdir(exist_ok=True)
    fig.savefig(file_path)
    fig.clf()


def single_plot(
    plot: plt.Axes,
    results: Results,
    xaxis: Tuple[str, str],
    yaxis: Tuple[str, str],
    dataset: str,
    transform: Optional[str],
    ptype: PlotType = "box",
    legend_pos: Optional[LegendType] = "outside",
    legend_yanchor: float = 1.0,
    markersize: int = 6,
    alternating_style: bool = True,
    include_nan_entries: bool = False,
) -> Union[None, Literal[False], legend.Legend]:
    """Provide the functionality of the individual plotting functions through a nice interface.

    This function can also be used to create figures with multiple plots on them, because it does
    not generate a Figure object itself.

    :param plot: Plot object
    :param results: DataFrame with the data
    :param xaxis: name of column that's plotted on the x-axis
    :param yaxis: name of column that's plotted on the y-axis
    :param dataset: string that identifies the dataset
    :param transform: string that identifies the preprocessing method, or None
    :param ptype: plot type (Default: "box")
    :param legend_pos: position of the legend (or None for no legend) (Default: "outside")
    :param legend_yanchor: position in the vertical direction where the legend should begin (Default
        value = 1.0)
    :param markersize: size of marker (Default: 6)
    :param alternating_style: if True, entries for scatter plots are done in alternating style
        (Default: True)
    :param include_nan_entries: if True, entries with NaNs still appear in the legend (Default value
        = False)
    :returns: the legend object if something was plotted; False otherwise
    """
    mask_for_dataset = results.index.get_level_values("dataset") == dataset
    if transform is not None:
        transforms: List[str] = [transform]
    else:
        transforms = [str(t) for t in results.index.to_frame()["transform"].unique()]

    entries: List[DataEntry] = []
    count = 0
    for model in results.index.to_frame()["model"].unique():
        mask_for_model = results.index.get_level_values("model") == model
        for transform_ in transforms:
            mask_for_transform = results.index.get_level_values("transform") == transform_
            data = results.loc[mask_for_dataset & mask_for_model & mask_for_transform]
            if (
                data[[xaxis[0], yaxis[0]]].empty or data[[xaxis[0], yaxis[0]]].isnull().any().any()
            ) and not include_nan_entries:
                continue  # this entry has missing values
            model_label = f"{model} ({transform_})" if transform_ != "no_transform" else str(model)
            entries.append(DataEntry(model_label, data, (not alternating_style) or count % 2 == 0))
            count += 1

    if not entries:
        return False  # nothing to plot

    title = f"{dataset}, {transform}" if transform is not None else dataset
    plot_def = PlotDef(
        title=title, entries=entries, legend_pos=legend_pos, legend_yanchor=legend_yanchor
    )
    if ptype == "box":
        return errorbox(plot, plot_def, xaxis, yaxis, 0, 0, markersize, use_cross=False)
    if ptype == "cross":
        return errorbox(plot, plot_def, xaxis, yaxis, 0, 0, markersize, use_cross=True)
    if ptype == "scatter":
        return scatter(plot, plot_def, xaxis, yaxis, 0, markersize, connect_dots=False)
    if ptype == "line":
        return scatter(plot, plot_def, xaxis, yaxis, 0, markersize, connect_dots=True)
    raise ValueError(f"Unknown plot type '{ptype}'")


def plot_results(
    results: Results,
    metric_y: Union[str, Metric],
    metric_x: Union[str, Metric],
    ptype: PlotType = "box",
    save: bool = True,
    dpi: int = 300,
    transforms_separately: bool = True,
) -> List[Tuple[figure.Figure, plt.Axes]]:
    """Plot the given result with boxes that represent mean and standard deviation.

    :param results: a DataFrame that already contains the values of the metrics
    :param metric_y: a Metric object or a column name that defines which metric to plot on the y-axis
    :param metric_x: a Metric object or a column name that defines which metric to plot on the x-axis
    :param ptype: plot type (Default: "box")
    :param save: if True, save the plot as a PDF (Default: True)
    :param dpi: DPI of the plots (Default: 300)
    :param transforms_separately: if True, each transform gets its own plot (Default: True)
    :returns: A list of all figures and plots
    """
    directory = Path(".") / "plots"
    directory.mkdir(exist_ok=True)

    def _get_columns(metric: Metric) -> List[str]:
        cols = [col for col in results.columns if metric.name in col]
        if not cols:
            raise ValueError(f'No matching columns found for Metric "{metric.name}".')
        # if there are multiple matches, then the metric was `per_sensitive_attribute`. In this
        # case, we *only* want ratios and differences; not the plain result
        if len(cols) > 1:
            cols = [col for col in cols if ("-" in col) or ("/" in col)]
        return cols

    if isinstance(metric_x, str):
        if metric_x not in results.columns:
            raise ValueError(f'No column named "{metric_x}".')
        cols_x = [metric_x]
    else:
        # if the metric is given as a Metric object, look for matching columns
        cols_x = _get_columns(metric_x)

    if isinstance(metric_y, str):
        if metric_y not in results.columns:
            raise ValueError(f'No column named "{metric_y}".')
        cols_y = [metric_y]
    else:
        cols_y = _get_columns(metric_y)

    # generate the Cartesian product of `cols_x` and `cols_y`; i.e. all possible combinations
    # this preserves the order of x and y
    possible_pairs = list(itertools.product(cols_x, cols_y))

    transforms: List[Optional[str]]
    if transforms_separately:
        transforms = [str(t) for t in results.index.to_frame()["transform"].unique()]
    else:
        transforms = [None]

    figure_list: List[Tuple[figure.Figure, plt.Axes]] = []
    for dataset in results.index.to_frame()["dataset"].unique():
        dataset_: str = str(dataset)
        for transform in transforms:
            for x_axis, y_axis in possible_pairs:
                fig: figure.Figure
                plot: plt.Axes
                fig, plot = plt.subplots(dpi=dpi)

                xtuple = (x_axis, x_axis.replace("_", " "))
                ytuple = (y_axis, y_axis.replace("_", " "))
                legend = single_plot(
                    plot, results, xtuple, ytuple, dataset_, transform, ptype=ptype
                )
                plt.tight_layout()

                if legend is False:  # use "is" here because otherwise any falsy value would match
                    plt.close(fig)
                    continue  # nothing was plotted -> don't save it and don't add it to the list

                if save:
                    metric_a = x_axis.replace("/", "_over_")
                    metric_b = y_axis.replace("/", "_over_")
                    fig.savefig(
                        directory / f"{dataset} {transform} {metric_a} {metric_b}.pdf",
                        bbox_inches="tight",
                    )
                plt.close(fig)
                figure_list += [(fig, plot)]
    return figure_list



"""Base class for Algorithms."""
import subprocess
import sys
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Dict, List, Optional

__all__ = ["Algorithm", "SubprocessAlgorithmMixin"]


class Algorithm(ABC):
    """Base class for Algorithms."""

    @property
    def name(self) -> str:
        """Name of the algorithm."""
        return self.get_name()

    # a method is nicer to implement than a property because you can use the @implements decorator
    @abstractmethod
    def get_name(self) -> str:
        """Name of the algorithm."""


class SubprocessAlgorithmMixin(ABC):  # pylint: disable=too-few-public-methods
    """Mixin for running algorithms in a subprocess, to be used with :class:`Algorithm`."""

    @property
    def executable(self) -> str:
        """Path to a (Python) executable.

        By default, the Python executable that called this script is used.
        """
        return sys.executable

    def call_script(
        self, cmd_args: List[str], env: Optional[Dict[str, str]] = None, cwd: Optional[Path] = None
    ) -> None:
        """Call a (Python) script as a separate process.

        An exception is thrown if the called script failed.

        :param cmd_args: List of strings that are passed as commandline arguments to the executable.
        :param env: Environment variables specified as a dictionary; e.g. ``{"PATH": "/usr/bin"}``.
        :param cwd: If not None, change working directory to the given path before running command.
        """
        two_hours = 60 * 60 * 2  # 60secs * 60mins * 2 hours
        try:
            process = subprocess.run(  # wait for process creation to finish
                [self.executable] + cmd_args,
                capture_output=True,
                env=env,
                cwd=cwd,
                timeout=two_hours,
            )
        except subprocess.TimeoutExpired as error:
            raise RuntimeError("The script timed out.") from error

        if process.returncode != 0:
            print(f"Failure: {cmd_args!r}")
            stderr = process.stderr
            if stderr:
                print(stderr.decode().strip())
            raise RuntimeError(
                f"The script failed. Supplied arguments: {cmd_args} with exec: {self.executable}"
            )



"""For assessing FNR."""
from dataclasses import dataclass
from typing import ClassVar

from ranzen import implements

from ethicml.utility import EvalTuple, Prediction

from .confusion_matrix import CfmMetric
from .metric import Metric

__all__ = ["FNR"]


@dataclass
class FNR(CfmMetric):
    """False negative rate."""

    _name: ClassVar[str] = "FNR"

    @implements(Metric)
    def score(self, prediction: Prediction, actual: EvalTuple) -> float:
        _, _, f_neg, t_pos = self.confusion_matrix(prediction=prediction, actual=actual)
        return f_neg / (f_neg + t_pos)



"""Simply returns a random (but legal) label."""
from __future__ import annotations

from dataclasses import dataclass
from typing import ClassVar

import numpy as np
import pandas as pd
from ranzen import implements

from ethicml.algorithms.inprocess.in_algorithm import InAlgorithmNoParams
from ethicml.utility import DataTuple, Prediction, TestTuple

__all__ = ["Blind"]


@dataclass
class Blind(InAlgorithmNoParams):
    """A Random classifier.

    This is useful as a baseline method and operates a 'coin flip' to assign a label.
    Returns a random label.
    """

    is_fairness_algo: ClassVar[bool] = False

    @implements(InAlgorithmNoParams)
    def get_name(self) -> str:
        return "Blind"

    @implements(InAlgorithmNoParams)
    def fit(self, train: DataTuple, seed: int = 888) -> "Blind":
        self.vals = train.y.drop_duplicates()
        self.seed = seed
        return self

    @implements(InAlgorithmNoParams)
    def predict(self, test: TestTuple) -> Prediction:
        random = np.random.RandomState(self.seed)
        return Prediction(hard=pd.Series(random.choice(self.vals.to_numpy(), test.x.shape[0])))

    @implements(InAlgorithmNoParams)
    def run(self, train: DataTuple, test: TestTuple, seed: int = 888) -> Prediction:
        vals = train.y.drop_duplicates()
        random = np.random.RandomState(seed)
        return Prediction(hard=pd.Series(random.choice(vals.to_numpy(), test.x.shape[0])))



"""For assessing Average Odds Difference metric."""
from dataclasses import dataclass
from typing import ClassVar

from ranzen import implements

from ethicml.utility import EvalTuple, Prediction

from .confusion_matrix import CfmMetric
from .fpr import FPR
from .metric import Metric
from .per_sensitive_attribute import diff_per_sensitive_attribute, metric_per_sensitive_attribute
from .tpr import TPR

__all__ = ["AverageOddsDiff"]


@dataclass
class AverageOddsDiff(CfmMetric):
    r"""Average Odds Difference.

    :math:`\tfrac{1}{2}\left[(FPR_{s=0} - FPR_{s=1}) + (TPR_{s=0} - TPR_{s=1}))\right]`.

    A value of 0 indicates equality of odds.
    """

    _name: ClassVar[str] = "AverageOddsDiff"
    apply_per_sensitive: ClassVar[bool] = False

    @implements(Metric)
    def score(self, prediction: Prediction, actual: EvalTuple) -> float:
        tpr = TPR(pos_class=self.pos_class, labels=self.labels)
        tpr_per_sens = metric_per_sensitive_attribute(prediction, actual, tpr)
        fpr = FPR(pos_class=self.pos_class, labels=self.labels)
        fpr_per_sens = metric_per_sensitive_attribute(prediction, actual, fpr)

        tpr_diff = diff_per_sensitive_attribute(tpr_per_sens)
        fpr_diff = diff_per_sensitive_attribute(fpr_per_sens)

        tpr_sum = 0.0
        fpr_sum = 0.0
        total = 0
        for ((_, tpr_v), (_, fpr_v)) in zip(tpr_diff.items(), fpr_diff.items()):
            total += 1
            tpr_sum += tpr_v
            fpr_sum += fpr_v

        return 0.5 * ((fpr_sum / total) + (tpr_sum / total))



"""Simple class to make class labels binary. Useful if a network uses BCELoss for example."""

import numpy as np
import pandas as pd

from ethicml.utility import DataTuple

__all__ = ["LabelBinarizer"]


def assert_binary_labels(data_tuple: DataTuple) -> None:
    """Assert that datasets only include binary labels.

    :param data_tuple:
    """
    assert data_tuple.y.nunique() == 2
    assert (np.unique(data_tuple.y.to_numpy()) == np.array([0, 1])).all()


class LabelBinarizer:
    """If a dataset has labels [-1,1], then this will make it so the labels = [0,1]."""

    def __init__(self) -> None:
        self.min_val: int
        self.max_val: int

    def adjust(self, dataset: DataTuple) -> DataTuple:
        """Take a datatuple and make the labels [0,1].

        :param dataset:
        """
        assert dataset.y.nunique() == 2

        # make copy of dataset
        new_y = dataset.y.copy()

        self.min_val = new_y.to_numpy().min().item()
        self.max_val = new_y.to_numpy().max().item()

        new_y = new_y.replace(self.min_val, 0)
        new_y = new_y.replace(self.max_val, 1)
        return dataset.replace(y=new_y)

    def post_only_labels(self, labels: pd.Series) -> pd.Series:
        """Inverse of adjust but only for a DataFrame instead of a DataTuple.

        :param labels:
        """
        assert labels.nunique() == 2

        # make copy of the labels
        labels_copy = labels.copy()

        labels_copy = labels_copy.replace(0, self.min_val)
        labels_copy = labels_copy.replace(1, self.max_val)
        return labels_copy

    def post(self, dataset: DataTuple) -> DataTuple:
        """Inverse of adjust.

        :param dataset:
        """
        transformed_y = self.post_only_labels(dataset.y)
        return dataset.replace(y=pd.Series(transformed_y, name=dataset.y.name))



"""FairDummies Models."""
from __future__ import annotations

import random
from typing import Callable, Tuple
from typing_extensions import Literal, Self

import numpy as np
import pandas as pd
import torch
import torch.nn as nn
import torch.optim as optim
from sklearn.preprocessing import StandardScaler
from torch.utils.data import DataLoader

from ... import DataTuple
from ..pytorch_common import DeepModel, DeepRegModel, LinearModel, PandasDataSet
from .utility_functions import density_estimation


def covariance_diff_biased(z: torch.Tensor, w: torch.Tensor, scale: float = 1.0) -> torch.Tensor:
    """Covariance Difference."""
    # Center X,Xk
    m_z = z - torch.mean(z, 0, keepdim=True)
    m_w = w - torch.mean(w, 0, keepdim=True)
    # Compute covariance matrices
    szz = torch.mm(torch.t(m_z), m_z) / m_z.shape[0]
    sww = torch.mm(torch.t(w), m_w) / m_w.shape[0]

    return (szz - sww).pow(2).sum() / scale  # Compute loss


# defining discriminator class (for regression)
class RegDiscriminator(nn.Module):
    """Regression Discriminator."""

    def __init__(self, inp, out=1):

        super().__init__()
        self.net = nn.Sequential(
            nn.Linear(inp, 10 * inp),
            nn.ReLU(inplace=True),
            nn.Linear(10 * inp, out),
            nn.Sigmoid(),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward."""
        return self.net(x)


# defining discriminator class (for classification)
class ClassDiscriminator(nn.Module):
    """Classification Discriminator."""

    def __init__(self, out_dim: int, n_y: int, n_hidden: int = 32):
        super().__init__()
        self.network = nn.Sequential(
            nn.Linear(n_y, n_hidden),
            nn.ReLU(),
            nn.Linear(n_hidden, n_hidden),
            nn.ReLU(),
            nn.Linear(n_hidden, n_hidden),
            nn.ReLU(),
            nn.Linear(n_hidden, out_dim),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward."""
        return torch.sigmoid(self.network(x))


def pretrain_adversary_fast_loader(
    dis: nn.Module,
    *,
    model: nn.Module,
    x: torch.Tensor,
    y: torch.Tensor,
    a: torch.Tensor,
    at: torch.Tensor,
    optimizer: torch.optim.Optimizer,
    criterion: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
    lambdas: torch.Tensor,
) -> nn.Module:
    """Pretrain adversary."""
    yhat = model(x).detach()
    dis.zero_grad()
    if len(yhat.size()) == 1:
        yhat = yhat.unsqueeze(dim=1)
    real = torch.cat((yhat, at, y), 1)
    fake = torch.cat((yhat, a, y), 1)
    in_dis = torch.cat((real, fake), 0)
    out_dis = dis(in_dis)
    labels = torch.cat((torch.ones(real.shape[0], 1), torch.zeros(fake.shape[0], 1)), 0)
    loss = (criterion(out_dis, labels) * lambdas).mean()
    loss.backward()
    optimizer.step()
    return dis


def pretrain_adversary(
    dis: nn.Module,
    *,
    model: nn.Module,
    data_loader: torch.utils.data.DataLoader,
    optimizer: torch.optim.Optimizer,
    criterion: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
    lambdas: torch.Tensor,
) -> nn.Module:
    """Pretrain adversary."""
    for x, a, y, at in data_loader:
        dis = pretrain_adversary_fast_loader(
            dis=dis,
            model=model,
            x=x,
            y=y,
            a=a,
            at=at,
            optimizer=optimizer,
            criterion=criterion,
            lambdas=lambdas,
        )
    return dis


def pretrain_classifier(
    model: nn.Module,
    *,
    data_loader: torch.utils.data.DataLoader,
    optimizer: torch.optim.Optimizer,
    criterion: nn.Module,
) -> nn.Module:
    """Pretrain classifier."""
    for x, _, y, _ in data_loader:
        model.zero_grad()
        yhat = model(x)
        loss = criterion(yhat, y.squeeze().long())
        loss.backward()
        optimizer.step()
    return model


def pretrain_regressor(
    model: nn.Module,
    *,
    data_loader: torch.utils.data.DataLoader,
    optimizer: torch.optim.Optimizer,
    criterion: nn.Module,
) -> nn.Module:
    """Pretrain regressor."""
    for x, _, y, _ in data_loader:
        model.zero_grad()
        yhat = model(x)
        loss = criterion(yhat.squeeze(), y.squeeze())
        loss.backward()
        optimizer.step()
    return model


def pretrain_regressor_fast_loader(
    model: nn.Module,
    *,
    x: torch.Tensor,
    y: torch.Tensor,
    optimizer: torch.optim.Optimizer,
    criterion: nn.Module,
) -> nn.Module:
    """Pretrain regressor."""
    model.zero_grad()
    yhat = model(x)
    loss = criterion(yhat.squeeze(), y.squeeze())
    loss.backward()
    optimizer.step()
    return model


def train_classifier(
    model: nn.Module,
    *,
    dis: nn.Module,
    data_loader: torch.utils.data.DataLoader,
    pred_loss: nn.Module,
    dis_loss: nn.Module,
    clf_optimizer: torch.optim.Optimizer,
    adv_optimizer: torch.optim.Optimizer,
    lambdas: torch.Tensor,
    second_moment_scaling: torch.Tensor,
    dis_steps: int,
    loss_steps: int,
    num_classes: int,
) -> Tuple[nn.Module, nn.Module]:
    """Train classifier."""
    # Train adversary
    for _ in range(dis_steps):
        for x, a, y, at in data_loader:
            yhat = model(x)
            dis.zero_grad()
            if len(yhat.size()) == 1:
                yhat = yhat.unsqueeze(dim=1)
            real = torch.cat((yhat, at, y), 1)
            fake = torch.cat((yhat, a, y), 1)
            in_dis = torch.cat((real, fake), 0)
            out_dis = dis(in_dis)
            labels = torch.cat((torch.ones(real.shape[0], 1), torch.zeros(fake.shape[0], 1)), 0)
            loss_adv = (dis_loss(out_dis, labels) * lambdas).mean()
            loss_adv.backward()
            adv_optimizer.step()

    # Train predictor
    for _ in range(loss_steps):
        for x, a, y, at in data_loader:
            yhat = model(x)
            if len(yhat.size()) == 1:
                yhat = yhat.unsqueeze(dim=1)

            y_one_hot = torch.zeros(len(y), num_classes).scatter_(1, y.long(), 1.0)
            fake_one_hot = torch.cat((yhat, a, y_one_hot), 1)
            real_one_hot = torch.cat((yhat, at, y_one_hot), 1)

            loss_second_moment = covariance_diff_biased(fake_one_hot, real_one_hot)

            fake = torch.cat((yhat, a, y), 1)
            real = torch.cat((yhat, at, y), 1)

            in_dis = torch.cat((real, fake), 0)
            out_dis = dis(in_dis)
            model.zero_grad()
            out_dis = dis(in_dis)
            labels = torch.cat((torch.zeros(real.shape[0], 1), torch.ones(fake.shape[0], 1)), 0)
            loss_adv = (dis_loss(out_dis, labels) * lambdas).mean()
            clf_loss = (1.0 - lambdas) * pred_loss(yhat, y.squeeze().long())
            clf_loss += (dis_loss(dis(in_dis), labels) * lambdas).mean()
            clf_loss += lambdas * second_moment_scaling * loss_second_moment
            clf_loss.backward()
            clf_optimizer.step()

            break

    return model, dis


def inner_train_adversary_regression(
    model: nn.Module,
    *,
    dis: nn.Module,
    x: torch.Tensor,
    y: torch.Tensor,
    a: torch.Tensor,
    at: torch.Tensor,
    pred_loss: nn.Module,
    dis_loss: nn.Module,
    clf_optimizer: torch.optim.Optimizer,
    adv_optimizer: torch.optim.Optimizer,
    lambdas: torch.Tensor,
    second_moment_scaling: torch.Tensor,
    dis_steps: int,
    loss_steps: int,
) -> nn.Module:
    """Inner train."""
    yhat = model(x)
    dis.zero_grad()
    if len(yhat.size()) == 1:
        yhat = yhat.unsqueeze(dim=1)
    real = torch.cat((yhat, at, y), 1)
    fake = torch.cat((yhat, a, y), 1)
    in_dis = torch.cat((real, fake), 0)
    out_dis = dis(in_dis)
    labels = torch.cat((torch.ones(real.shape[0], 1), torch.zeros(fake.shape[0], 1)), 0)
    loss_adv = (dis_loss(out_dis, labels) * lambdas).mean()
    loss_adv.backward()
    adv_optimizer.step()
    return dis


def inner_train_model_regression(
    model: nn.Module,
    *,
    dis: nn.Module,
    x: torch.Tensor,
    y: torch.Tensor,
    a: torch.Tensor,
    at: torch.Tensor,
    pred_loss: nn.Module,
    dis_loss: nn.Module,
    clf_optimizer: torch.optim.Optimizer,
    adv_optimizer: torch.optim.Optimizer,
    lambdas: torch.Tensor,
    second_moment_scaling: torch.Tensor,
    dis_steps: int,
    loss_steps: int,
) -> nn.Module:
    """Inner train."""
    yhat = model(x)
    if len(yhat.size()) == 1:
        yhat = yhat.unsqueeze(dim=1)

    fake = torch.cat((yhat, a, y), 1)
    real = torch.cat((yhat, at, y), 1)

    loss_second_moment = covariance_diff_biased(fake, real)

    in_dis = torch.cat((real, fake), 0)
    out_dis = dis(in_dis)
    model.zero_grad()
    out_dis = dis(in_dis)
    labels = torch.cat((torch.zeros(real.shape[0], 1), torch.ones(fake.shape[0], 1)), 0)
    (dis_loss(out_dis, labels) * lambdas).mean()
    clf_loss = (1.0 - lambdas) * pred_loss(yhat.squeeze(), y.squeeze())
    clf_loss += (dis_loss(dis(in_dis), labels) * lambdas).mean()
    clf_loss += lambdas * second_moment_scaling * loss_second_moment
    clf_loss.backward()
    clf_optimizer.step()
    return model


def train_regressor_fast_loader(
    model: nn.Module,
    *,
    dis: nn.Module,
    x: torch.Tensor,
    y: torch.Tensor,
    a: torch.Tensor,
    at: torch.Tensor,
    pred_loss: nn.Module,
    dis_loss: nn.Module,
    clf_optimizer: torch.optim.Optimizer,
    adv_optimizer: torch.optim.Optimizer,
    lambdas: torch.Tensor,
    second_moment_scaling: torch.Tensor,
    dis_steps: int,
    loss_steps: int,
) -> Tuple[nn.Module, nn.Module]:
    """Train regressor."""
    # Train adversary
    for _ in range(dis_steps):
        dis = inner_train_adversary_regression(
            model=model,
            dis=dis,
            x=x,
            y=y,
            a=a,
            at=at,
            pred_loss=pred_loss,
            dis_loss=dis_loss,
            clf_optimizer=clf_optimizer,
            adv_optimizer=adv_optimizer,
            lambdas=lambdas,
            second_moment_scaling=second_moment_scaling,
            dis_steps=dis_steps,
            loss_steps=loss_steps,
        )

    # Train predictor
    for _ in range(loss_steps):
        model = inner_train_model_regression(
            model=model,
            dis=dis,
            x=x,
            y=y,
            a=a,
            at=at,
            pred_loss=pred_loss,
            dis_loss=dis_loss,
            clf_optimizer=clf_optimizer,
            adv_optimizer=adv_optimizer,
            lambdas=lambdas,
            second_moment_scaling=second_moment_scaling,
            dis_steps=dis_steps,
            loss_steps=loss_steps,
        )

    return model, dis


def train_regressor(
    model: nn.Module,
    *,
    dis: nn.Module,
    data_loader: torch.utils.data.DataLoader,
    pred_loss: nn.Module,
    dis_loss: nn.Module,
    clf_optimizer: torch.optim.Optimizer,
    adv_optimizer: torch.optim.Optimizer,
    lambdas: torch.Tensor,
    second_moment_scaling: torch.Tensor,
    dis_steps: int,
    loss_steps: int,
) -> Tuple[nn.Module, nn.Module]:
    """Train the regressor."""
    # Train adversary
    for _ in range(dis_steps):
        for x, a, y, at in data_loader:
            dis = inner_train_adversary_regression(
                model=model,
                dis=dis,
                x=x,
                y=y,
                a=a,
                at=at,
                pred_loss=pred_loss,
                dis_loss=dis_loss,
                clf_optimizer=clf_optimizer,
                adv_optimizer=adv_optimizer,
                lambdas=lambdas,
                second_moment_scaling=second_moment_scaling,
                dis_steps=dis_steps,
                loss_steps=loss_steps,
            )

    # Train predictor
    for _ in range(loss_steps):
        for x, a, y, at in data_loader:
            model = inner_train_model_regression(
                model=model,
                dis=dis,
                x=x,
                y=y,
                a=a,
                at=at,
                pred_loss=pred_loss,
                dis_loss=dis_loss,
                clf_optimizer=clf_optimizer,
                adv_optimizer=adv_optimizer,
                lambdas=lambdas,
                second_moment_scaling=second_moment_scaling,
                dis_steps=dis_steps,
                loss_steps=loss_steps,
            )

    return model, dis


def seed_worker(worker_id):
    """Seed the Dataloader worker."""
    worker_seed = torch.initial_seed() % 2**32
    np.random.seed(worker_seed)
    random.seed(worker_seed)


class EquiClassLearner:
    """Classification model."""

    def __init__(
        self,
        lr: float,
        pretrain_pred_epochs: int,
        pretrain_dis_epochs: int,
        epochs: int,
        loss_steps: int,
        dis_steps: int,
        cost_pred: nn.Module,
        in_shape: int,
        batch_size: int,
        model_type: Literal["deep_model", "linear_model"],
        lambda_vec: float,
        second_moment_scaling: float,
        num_classes: int,
        seed: int,
    ):
        random.seed(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)
        if torch.cuda.is_available():
            torch.cuda.manual_seed_all(seed)

        torch.use_deterministic_algorithms(True)

        self.seed = seed
        self.lr = lr
        self.batch_size = batch_size
        self.in_shape = in_shape
        self.num_classes = num_classes

        self.model_type = model_type
        if self.model_type == "deep_model":
            self.model: nn.Module = DeepModel(in_shape=in_shape, out_shape=num_classes)
        elif self.model_type == "linear_model":
            self.model = LinearModel(in_shape=in_shape, out_shape=num_classes)
        else:
            raise NotImplementedError

        self.pred_loss = cost_pred
        self.clf_optimizer = optim.Adam(self.model.parameters(), lr=self.lr)

        self.pretrain_pred_epochs = pretrain_pred_epochs

        self.lambdas = torch.Tensor([lambda_vec])
        self.second_moment_scaling = torch.Tensor([second_moment_scaling])

        self.dis: nn.Module = ClassDiscriminator(out_dim=1, n_y=num_classes + 1 + 1)
        self.dis_loss = nn.BCELoss(reduce=False)
        self.adv_optimizer = optim.Adam(self.dis.parameters(), lr=self.lr)

        self.pretrain_dis_epochs = pretrain_dis_epochs

        self.epochs = epochs
        self.loss_steps = loss_steps
        self.dis_steps = dis_steps

        self.scaler = StandardScaler()

    def fit(self, train: DataTuple, seed: int) -> Self:  # type: ignore[valid-type]
        """Fit."""
        # The features are X[:,1:]
        p_success, dummy = density_estimation(y=train.y.to_numpy(), a=train.s.to_numpy())

        rng = np.random.default_rng(seed)

        g = torch.Generator()
        g.manual_seed(self.seed)

        for _ in range(self.pretrain_pred_epochs):
            random_array = rng.uniform(low=0.0, high=1.0, size=train.s.shape)
            z_tilde = (random_array < p_success).astype(float)
            zt_train = pd.DataFrame(data=z_tilde)
            train_data = PandasDataSet(train.x, train.s, train.y, zt_train)
            train_loader = DataLoader(
                train_data,
                batch_size=self.batch_size,
                shuffle=True,
                drop_last=True,
                worker_init_fn=seed_worker,
                generator=g,
            )

            self.model = pretrain_classifier(
                model=self.model,
                data_loader=train_loader,
                optimizer=self.clf_optimizer,
                criterion=self.pred_loss,
            )

        for _ in range(self.pretrain_dis_epochs):
            random_array = rng.uniform(low=0.0, high=1.0, size=train.s.shape)
            z_tilde = (random_array < p_success).astype(float)
            zt_train = pd.DataFrame(data=z_tilde)
            train_data = PandasDataSet(train.x, train.s, train.y, zt_train)
            train_loader = DataLoader(
                train_data,
                batch_size=self.batch_size,
                shuffle=True,
                drop_last=True,
                worker_init_fn=seed_worker,
                generator=g,
            )

            pretrain_adversary(
                dis=self.dis,
                model=self.model,
                data_loader=train_loader,
                optimizer=self.adv_optimizer,
                criterion=self.dis_loss,
                lambdas=self.lambdas,
            )

        for _ in range(1, self.epochs):
            random_array = rng.uniform(low=0.0, high=1.0, size=train.s.shape)
            z_tilde = (random_array < p_success).astype(float)
            zt_train = pd.DataFrame(data=z_tilde)
            train_data = PandasDataSet(train.x, train.s, train.y, zt_train)
            train_loader = DataLoader(
                train_data,
                batch_size=self.batch_size,
                shuffle=True,
                drop_last=True,
                worker_init_fn=seed_worker,
                generator=g,
            )

            self.model, self.dis = train_classifier(
                model=self.model,
                dis=self.dis,
                data_loader=train_loader,
                pred_loss=self.pred_loss,
                dis_loss=self.dis_loss,
                clf_optimizer=self.clf_optimizer,
                adv_optimizer=self.adv_optimizer,
                lambdas=self.lambdas,
                second_moment_scaling=self.second_moment_scaling,
                dis_steps=self.dis_steps,
                loss_steps=self.loss_steps,
                num_classes=self.num_classes,
            )
        return self

    def predict(self, x: pd.DataFrame) -> np.ndarray:
        """Predict."""
        test_data = PandasDataSet(x)

        self.model.eval()
        with torch.no_grad():
            yhat = self.model(test_data.tensors[0])

        sm = nn.Softmax(dim=1)
        yhat = sm(yhat)
        yhat = yhat.detach().numpy()

        return yhat


class EquiRegLearner:
    """Regression model."""

    def __init__(
        self,
        lr: float,
        pretrain_pred_epochs: int,
        pretrain_dis_epochs: int,
        epochs: int,
        loss_steps: int,
        dis_steps: int,
        cost_pred: nn.Module,
        in_shape: int,
        batch_size: int,
        model_type: Literal["deep_model", "linear_model"],
        lambda_vec: float,
        second_moment_scaling: float,
        out_shape: int,
        use_standardscaler: bool = True,
    ):

        self.lr = lr
        self.batch_size = batch_size
        self.in_shape = in_shape
        self.out_shape = out_shape
        self.use_standardscaler = use_standardscaler

        self.model_type = model_type
        if self.model_type == "deep_model":
            self.model: nn.Module = DeepRegModel(in_shape=in_shape, out_shape=out_shape)
        elif self.model_type == "linear_model":
            self.model = LinearModel(in_shape=in_shape, out_shape=out_shape)
        else:
            raise NotImplementedError

        self.pred_loss = cost_pred
        self.clf_optimizer = torch.optim.SGD(self.model.parameters(), lr=self.lr)
        self.pretrain_pred_epochs = pretrain_pred_epochs

        self.lambdas = torch.Tensor([lambda_vec])
        self.second_moment_scaling = torch.Tensor([second_moment_scaling])

        self.dis: nn.Module = RegDiscriminator(out_shape + 1 + 1)
        self.dis_loss = nn.BCELoss(reduce=False)
        self.adv_optimizer = torch.optim.SGD(self.dis.parameters(), lr=self.lr, momentum=0.9)
        self.pretrain_dis_epochs = pretrain_dis_epochs

        self.epochs = epochs
        self.loss_steps = loss_steps
        self.dis_steps = dis_steps

    def fit(self, train: DataTuple, seed: int) -> None:
        """Fit."""
        fast_loader = self.batch_size >= train.x.shape[0]
        p_success, dummy = density_estimation(y=train.y.to_numpy(), a=train.s.to_numpy())

        x = torch.from_numpy(train.x.values).float()
        y = torch.from_numpy(train.y.values).float()
        a = torch.from_numpy(train.s.values).float()

        rng = np.random.default_rng(seed)
        g = torch.Generator()
        g.manual_seed(seed)

        for _ in range(self.pretrain_pred_epochs):
            random_array = rng.uniform(low=0.0, high=1.0, size=train.s.shape)
            z_tilde = (random_array < p_success).astype(float)
            zt_train = pd.DataFrame(data=z_tilde)
            train_data = PandasDataSet(train.x, train.s, train.y, zt_train)
            train_loader = DataLoader(
                train_data,
                batch_size=self.batch_size,
                shuffle=True,
                drop_last=False,
                worker_init_fn=seed_worker,
                generator=g,
            )
            self.model = (
                pretrain_regressor_fast_loader(
                    model=self.model,
                    x=x,
                    y=y,
                    optimizer=self.clf_optimizer,
                    criterion=self.pred_loss,
                )
                if fast_loader
                else pretrain_regressor(
                    model=self.model,
                    data_loader=train_loader,
                    optimizer=self.clf_optimizer,
                    criterion=self.pred_loss,
                )
            )

        for _ in range(self.pretrain_dis_epochs):
            random_array = rng.uniform(low=0.0, high=1.0, size=train.s.shape)
            z_tilde = (random_array < p_success).astype(float)
            zt_train = pd.DataFrame(data=z_tilde)
            train_data = PandasDataSet(train.x, train.s, train.y, zt_train)
            train_loader = DataLoader(
                train_data,
                batch_size=self.batch_size,
                shuffle=True,
                drop_last=False,
                worker_init_fn=seed_worker,
                generator=g,
            )
            if fast_loader:
                pretrain_adversary_fast_loader(
                    dis=self.dis,
                    model=self.model,
                    x=x,
                    y=y,
                    a=a,
                    at=torch.from_numpy(zt_train.values).float(),
                    optimizer=self.adv_optimizer,
                    criterion=self.dis_loss,
                    lambdas=self.lambdas,
                )
            else:
                pretrain_adversary(
                    dis=self.dis,
                    model=self.model,
                    data_loader=train_loader,
                    optimizer=self.adv_optimizer,
                    criterion=self.dis_loss,
                    lambdas=self.lambdas,
                )

        for _ in range(1, self.epochs):
            random_array = rng.uniform(low=0.0, high=1.0, size=train.s.shape)
            z_tilde = (random_array < p_success).astype(float)
            zt_train = pd.DataFrame(data=z_tilde)
            train_data = PandasDataSet(train.x, train.s, train.y, zt_train)
            train_loader = DataLoader(
                train_data,
                batch_size=self.batch_size,
                shuffle=True,
                drop_last=False,
                worker_init_fn=seed_worker,
                generator=g,
            )
            self.model, self.dis = (
                train_regressor_fast_loader(
                    model=self.model,
                    dis=self.dis,
                    x=x,
                    y=y,
                    a=a,
                    at=torch.from_numpy(zt_train.values).float(),
                    pred_loss=self.pred_loss,
                    dis_loss=self.dis_loss,
                    clf_optimizer=self.clf_optimizer,
                    adv_optimizer=self.adv_optimizer,
                    lambdas=self.lambdas,
                    second_moment_scaling=self.second_moment_scaling,
                    dis_steps=self.dis_steps,
                    loss_steps=self.loss_steps,
                )
                if fast_loader
                else train_regressor(
                    model=self.model,
                    dis=self.dis,
                    data_loader=train_loader,
                    pred_loss=self.pred_loss,
                    dis_loss=self.dis_loss,
                    clf_optimizer=self.clf_optimizer,
                    adv_optimizer=self.adv_optimizer,
                    lambdas=self.lambdas,
                    second_moment_scaling=self.second_moment_scaling,
                    dis_steps=self.dis_steps,
                    loss_steps=self.loss_steps,
                )
            )

    def predict(self, x: pd.DataFrame) -> float:
        """Predict."""
        test_data = PandasDataSet(x)

        self.model.eval()
        with torch.no_grad():
            yhat = self.model(test_data.tensors[0]).squeeze().detach().numpy()

        if self.out_shape == 1:
            out = yhat.squeeze()
        else:
            out = 0 * yhat
            out[:, 0] = np.min(yhat, axis=1)
            out[:, 1] = np.max(yhat, axis=1)

        return out



"""For assessing Balanced Classification Rate (BCR)."""
from dataclasses import dataclass
from typing import ClassVar

from ranzen import implements

from ethicml.utility import EvalTuple, Prediction

from .confusion_matrix import CfmMetric
from .metric import Metric
from .tnr import TNR
from .tpr import TPR

__all__ = ["BCR"]


@dataclass
class BCR(CfmMetric):
    """Balanced Classification Rate."""

    _name: ClassVar[str] = "BCR"

    @implements(Metric)
    def score(self, prediction: Prediction, actual: EvalTuple) -> float:
        tpr = TPR(pos_class=self.pos_class, labels=self.labels).score(prediction, actual)
        tnr = TNR(pos_class=self.pos_class, labels=self.labels).score(prediction, actual)

        return (tpr + tnr) / 2



"""HGR."""
# Code follows the implemention provided in
# https://github.com/criteo-research/continuous-fairness
# The function for measuring HGR is in the facl package, can be downloaded from
# https://github.com/criteo-research/continuous-fairness/tree/master/facl/independence
import random
from typing_extensions import Literal, Self

import numpy as np
import pandas as pd
import torch
import torch.nn as nn
from sklearn.preprocessing import StandardScaler

from ... import DataTuple
from ..pytorch_common import DeepModel, DeepRegModel, LinearModel, make_dataset_and_loader
from .density_estimation import Kde
from .facl_hgr import chi_2_cond


def chi_squared_l1_kde(x, y, z):
    """Chi Squared."""
    return torch.mean(chi_2_cond(x, y, z, Kde))


class HgrRegLearner:
    """HGR Regression."""

    def __init__(self, lr, epochs, mu, cost_pred, in_shape, out_shape, batch_size, model_type):

        self.in_shape = in_shape
        self.model_type = model_type

        # Data normalization
        self.x_scaler = StandardScaler()
        self.a_scaler = StandardScaler()
        self.y_scaler = StandardScaler()

        # EO penalty
        self.mu = mu

        # Loss optimization
        self.cost_pred = cost_pred
        self.epochs = epochs
        self.lr_loss = lr
        self.batch_size = batch_size

        self.out_shape = out_shape
        if self.model_type == "deep_model":
            self.model: nn.Module = DeepRegModel(in_shape=in_shape, out_shape=out_shape)
        elif self.model_type == "linear_model":
            self.model = LinearModel(in_shape=in_shape, out_shape=out_shape)

        else:
            raise NotImplementedError
        self.loss_optimizer = torch.optim.Adam(self.model.parameters(), lr=self.lr_loss)

    def internal_epoch(self, dataloader: torch.utils.data.DataLoader) -> np.ndarray:
        """Internal epochs."""
        # fit pred func
        epoch_losses = []
        for x, s, y in dataloader:
            self.loss_optimizer.zero_grad()

            batch_yhat = self.model(x)

            # utility loss
            pred_loss = self.cost_pred(batch_yhat, y)

            if self.out_shape == 1:
                dis_loss = chi_squared_l1_kde(batch_yhat, s, y)
            else:
                dis_loss = sum(
                    chi_squared_l1_kde(batch_yhat[:, out_id], s, y)
                    for out_id in range(batch_yhat.shape[1])
                )

            # combine utility + 'distance to equalized odds'
            loss = (1 - self.mu) * pred_loss + self.mu * dis_loss

            loss.backward()
            self.loss_optimizer.step()

            epoch_losses.append(loss.detach().cpu().numpy())

        return np.mean(epoch_losses)

    def run_epochs(self, dataloader: torch.utils.data.DataLoader) -> None:
        """Run Epochs."""
        for _ in range(self.epochs):
            self.internal_epoch(dataloader)

    def fit(self, train: DataTuple, seed: int) -> None:
        """Fit."""
        torch.use_deterministic_algorithms(True)
        torch.manual_seed(seed)
        random.seed(seed)
        np.random.seed(seed)
        train_data, train_loader = make_dataset_and_loader(
            train, batch_size=self.batch_size, shuffle=True, seed=seed, drop_last=True
        )
        self.run_epochs(train_loader)

    @torch.no_grad()
    def predict(self, x: pd.DataFrame) -> np.ndarray:
        """Predict."""
        xp = torch.from_numpy(self.x_scaler.transform(x.to_numpy())).float()
        yhat = self.model(xp).detach()
        yhat = yhat.detach().cpu().numpy()

        if self.out_shape == 1:
            out = self.y_scaler.inverse_transform(yhat.reshape(-1, 1)).squeeze()
        else:
            out = 0 * yhat
            out[:, 0] = np.min(yhat, axis=1)
            out[:, 1] = np.max(yhat, axis=1)

        return out


class HgrClassLearner:
    """HGR Class."""

    def __init__(
        self,
        lr: float,
        epochs: int,
        mu: float,
        cost_pred: nn.Module,
        in_shape: int,
        out_shape: int,
        batch_size: int,
        model_type: Literal["deep_model", "linear_model"],
    ):

        self.in_shape = in_shape
        self.num_classes = out_shape

        # EO penalty
        self.mu = mu

        # Loss optimization
        self.cost_pred = cost_pred
        self.epochs = epochs
        self.lr_loss = lr
        self.batch_size = batch_size

        self.model_type = model_type
        if self.model_type == "deep_model":
            self.model: nn.Module = DeepModel(in_shape=in_shape, out_shape=out_shape)
        elif self.model_type == "linear_model":
            self.model = LinearModel(in_shape=in_shape, out_shape=out_shape)
        else:
            raise NotImplementedError

        self.loss_optimizer = torch.optim.Adam(self.model.parameters(), lr=self.lr_loss)

    def internal_epoch(self, dataloader: torch.utils.data.DataLoader) -> None:
        """Internal Epoch."""
        # fit pred func
        epoch_losses = []
        for x, s, y in dataloader:
            self.loss_optimizer.zero_grad()
            batch_yhat = self.model(x)

            # utility loss
            pred_loss = self.cost_pred(batch_yhat, y.long())

            dis_loss = sum(
                chi_squared_l1_kde(batch_yhat[:, out_id], s, y.float())
                for out_id in range(batch_yhat.shape[1])
            )

            # combine utility + 'distance to equalized odds'
            loss = (1 - self.mu) * pred_loss + self.mu * dis_loss

            loss.backward()
            self.loss_optimizer.step()

            epoch_losses.append(loss.cpu().detach().numpy())

    def run_epochs(self, dataloader: torch.utils.data.DataLoader) -> None:
        """Run epochs."""
        for _ in range(self.epochs):
            self.internal_epoch(dataloader)

    def fit(self, train: DataTuple, seed: int) -> Self:  # type: ignore[valid-type]
        """Fit."""
        torch.manual_seed(seed)
        random.seed(seed)
        np.random.seed(seed)
        torch.use_deterministic_algorithms(True)
        # train
        train_data, train_loader = make_dataset_and_loader(
            train, batch_size=self.batch_size, shuffle=True, seed=seed, drop_last=True
        )
        self.model.train()
        self.run_epochs(train_loader)
        return self

    @torch.no_grad()
    def predict(self, x: pd.DataFrame) -> np.ndarray:
        """Predict."""
        xp = torch.from_numpy(x.to_numpy()).float()
        self.model.eval()
        yhat = self.model(xp)
        sm = nn.Softmax(dim=1)
        yhat = sm(yhat)
        return yhat.detach().numpy()



"""Beutel's algorithm."""
from dataclasses import dataclass, field
from typing import List
from typing_extensions import TypedDict

from ranzen import implements

from ethicml.utility import FairnessType

from .pre_subprocess import PreAlgorithmSubprocess

__all__ = ["Beutel"]


class BeutelArgs(TypedDict):
    """Args for the Beutel Implementation."""

    fairness: str
    enc_size: List[int]
    adv_size: List[int]
    pred_size: List[int]
    enc_activation: str
    adv_activation: str
    batch_size: int
    y_loss: str
    s_loss: str
    epochs: int
    adv_weight: float
    validation_pcnt: float


@dataclass
class Beutel(PreAlgorithmSubprocess):
    """Beutel's adversarially learned fair representations."""

    fairness: FairnessType = FairnessType.dp
    enc_size: List[int] = field(default_factory=lambda: [40])
    adv_size: List[int] = field(default_factory=lambda: [40])
    pred_size: List[int] = field(default_factory=lambda: [40])
    enc_activation: str = "Sigmoid()"
    adv_activation: str = "Sigmoid()"
    batch_size: int = 64
    y_loss: str = "CrossEntropyLoss()"
    s_loss: str = "BCELoss()"
    epochs: int = 50
    adv_weight: float = 1.0
    validation_pcnt: float = 0.1

    @implements(PreAlgorithmSubprocess)
    def get_out_size(self) -> int:
        return self.enc_size[-1]

    @implements(PreAlgorithmSubprocess)
    def _get_flags(self) -> BeutelArgs:
        # TODO: replace this with dataclasses.asdict()
        return {
            "fairness": str(self.fairness),
            "enc_size": list(self.enc_size),
            "adv_size": list(self.adv_size),
            "pred_size": list(self.pred_size),
            "enc_activation": self.enc_activation,
            "adv_activation": self.adv_activation,
            "batch_size": self.batch_size,
            "y_loss": self.y_loss,
            "s_loss": self.s_loss,
            "epochs": self.epochs,
            "adv_weight": self.adv_weight,
            "validation_pcnt": self.validation_pcnt,
        }

    @implements(PreAlgorithmSubprocess)
    def get_name(self) -> str:
        return f"Beutel {self.fairness}"

    @implements(PreAlgorithmSubprocess)
    def _get_path_to_script(self) -> List[str]:
        return ["-m", "ethicml.implementations.beutel"]



"""Fairness without Demographics Classifier."""
from typing import List

from ranzen import implements
from ranzen.torch import CrossEntropyLoss
from torch import Tensor, nn

from .dro_loss import DROLoss

__all__ = ["DROClassifier"]


class DROClassifier(nn.Module):
    """Simple Classifier using Fairness without Demographics Loss."""

    def __init__(self, in_size: int, out_size: int, network_size: List[int], eta: float) -> None:
        super().__init__()

        self.loss = DROLoss(loss_module=CrossEntropyLoss, eta=eta)

        self.seq = nn.Sequential()
        if not network_size:  # In the case that encoder size [] is specified
            self.seq.add_module("DRO Model single layer", nn.Linear(in_size, out_size))
        else:
            self.seq.add_module("DRO Model layer 0", nn.Linear(in_size, network_size[0]))
            self.seq.add_module("DRO Model activation 0", nn.ReLU())
            self.seq.add_module("DRO Model batch norm 0", nn.BatchNorm1d(network_size[0]))
            for k in range(len(network_size) - 1):
                self.seq.add_module(
                    f"DRO Model layer {k + 1}", nn.Linear(network_size[k], network_size[k + 1])
                )
                self.seq.add_module(f"DRO Model activation {k + 1}", nn.ReLU())
                self.seq.add_module(
                    f"DRO Model batch norm {k + 1}", nn.BatchNorm1d(network_size[k + 1])
                )
            self.seq.add_module("DRO Model last layer", nn.Linear(network_size[-1], out_size))

    @implements(nn.Module)
    def forward(self, x: Tensor) -> Tensor:
        return self.seq(x)



"""Implementation of Heaviside decision function."""

import numpy

from .activation import Activation

__all__ = ["Heaviside"]


class Heaviside(Activation):
    """Decision function that accepts predictions with score of 50% or above."""

    def apply(self, soft_output: numpy.ndarray) -> numpy.ndarray:
        """Apply the decision function to each element of an ndarray.

        :param soft_output: Soft predictions.
        """

        def _heavi(x: float) -> int:
            return 1 if x >= 0.5 else 0

        return numpy.array([_heavi(x) for x in soft_output])

    def get_name(self) -> str:
        """Getter for name of decision function."""
        return "Heaviside"



"""Independence of 2 variables."""
from typing_extensions import Type

import numpy as np
import torch

from .density_estimation import Kde


def _joint_2(
    x: torch.Tensor, y: torch.Tensor, density: Type[Kde], damping: float = 1e-10
) -> torch.Tensor:
    x = (x - x.mean()) / x.std()
    y = (y - y.mean()) / y.std()
    data = torch.cat([x.unsqueeze(-1), y.unsqueeze(-1)], -1)
    joint_density = density(data)

    nbins = int(min(50, 5.0 / joint_density.std))
    # nbins = np.sqrt( Y.size/5 )
    x_centers = torch.linspace(-2.5, 2.5, nbins)
    y_centers = torch.linspace(-2.5, 2.5, nbins)

    xx, yy = torch.meshgrid([x_centers, y_centers])
    grid = torch.cat([xx.unsqueeze(-1), yy.unsqueeze(-1)], -1)
    h2d: torch.Tensor = joint_density.pdf(grid) + damping
    h2d /= h2d.sum()
    return h2d


def hgr(x: torch.Tensor, y: torch.Tensor, density: Type[Kde], damping: float = 1e-10) -> float:
    """An estimator of the Hirschfeld-Gebelein-Renyi maximum correlation coefficient using Witsenhausen’s Characterization.

    HGR(x,y) is the second highest eigenvalue of the joint density on (x,y). We compute here the second eigenvalue on
    an empirical and discretized density estimated from the input data.
    :param x: A torch 1-D Tensor
    :param y: A torch 1-D Tensor
    :param density: so far only kde is supported
    :return: numerical value between 0 and 1 (0: independent, 1:linked by a deterministic equation)
    """
    h2d = _joint_2(x, y, density, damping=damping)
    marginal_x = h2d.sum(dim=1).unsqueeze(1)
    marginal_y = h2d.sum(dim=0).unsqueeze(0)
    Q = h2d / (torch.sqrt(marginal_x) * torch.sqrt(marginal_y))
    return torch.svd(Q)[1][1]


def chi_2(x: torch.Tensor, y: torch.Tensor, density: Type[Kde], damping: float = 0) -> float:
    r"""The \chi^2 divergence between the joint distribution on (x,y) and the product of marginals.

    This is know to be the square of an upper-bound on the Hirschfeld-Gebelein-Renyi maximum correlation coefficient. We compute it here on
    an empirical and discretized density estimated from the input data.
    :param x: A torch 1-D Tensor
    :param y: A torch 1-D Tensor
    :param density: so far only kde is supported
    :return: numerical value between 0 and \infty (0: independent)
    """
    h2d = _joint_2(x, y, density, damping=damping)
    marginal_x = h2d.sum(dim=1).unsqueeze(1)
    marginal_y = h2d.sum(dim=0).unsqueeze(0)
    Q = h2d / (torch.sqrt(marginal_x) * torch.sqrt(marginal_y))
    return (Q**2).sum(dim=[0, 1]) - 1.0


# Independence of conditional variables


def _joint_3(
    x: torch.Tensor, y: torch.Tensor, z: torch.Tensor, density: Type[Kde], damping: float = 1e-10
) -> torch.Tensor:
    x = (x - x.mean()) / x.std()
    y = (y - y.mean()) / y.std()
    z = (z - z.mean()) / z.std()
    data = torch.cat([x.unsqueeze(-1), y.unsqueeze(-1), z.unsqueeze(-1)], -1)
    joint_density = density(data)  # + damping

    nbins = int(min(50, 5.0 / joint_density.std))
    x_centers = torch.linspace(-2.5, 2.5, nbins)
    y_centers = torch.linspace(-2.5, 2.5, nbins)
    z_centers = torch.linspace(-2.5, 2.5, nbins)
    xx, yy, zz = torch.meshgrid([x_centers, y_centers, z_centers])
    grid = torch.cat([xx.unsqueeze(-1), yy.unsqueeze(-1), zz.unsqueeze(-1)], -1)

    h3d: torch.Tensor = joint_density.pdf(grid) + damping
    h3d /= h3d.sum()
    return h3d


def hgr_cond(x: torch.Tensor, y: torch.Tensor, z: torch.Tensor, density: Type[Kde]) -> np.ndarray:
    """An estimator of the function z -> HGR(x|z, y|z).

    Where HGR is the Hirschfeld-Gebelein-Renyi maximum correlation
    coefficient computed using Witsenhausen’s Characterization: HGR(x,y) is the second highest eigenvalue of the joint
    density on (x,y). We compute here the second eigenvalue on
    an empirical and discretized density estimated from the input data.
    :param x: A torch 1-D Tensor
    :param y: A torch 1-D Tensor
    :param z: A torch 1-D Tensor
    :param density: so far only kde is supported
    :return: A torch 1-D Tensor of same size as Z. (0: independent, 1:linked by a deterministic equation)
    """
    damping = 1e-10
    h3d = _joint_3(x, y, z, density, damping=damping)
    marginal_xz = h3d.sum(dim=1).unsqueeze(1)
    marginal_yz = h3d.sum(dim=0).unsqueeze(0)
    Q = h3d / (torch.sqrt(marginal_xz) * torch.sqrt(marginal_yz))
    return np.array(([torch.svd(Q[:, :, i])[1][1] for i in range(Q.shape[2])]))


def chi_2_cond(
    x: torch.Tensor, y: torch.Tensor, z: torch.Tensor, density: Type[Kde]
) -> torch.Tensor:
    r"""An estimator of the function z -> chi^2(x|z, y|z).

    Where \chi^2 is the \chi^2 divergence between the joint
    distribution on (x,y) and the product of marginals. This is know to be the square of an upper-bound on the
    Hirschfeld-Gebelein-Renyi maximum correlation coefficient. We compute it here on an empirical and discretized
    density estimated from the input data.
    :param x: A torch 1-D Tensor
    :param y: A torch 1-D Tensor
    :param z: A torch 1-D Tensor
    :param density: so far only kde is supported
    :return: A torch 1-D Tensor of same size as Z. (0: independent)
    """
    damping = 0
    h3d = _joint_3(x, y, z, density, damping=damping)
    marginal_xz = h3d.sum(dim=1).unsqueeze(1)
    marginal_yz = h3d.sum(dim=0).unsqueeze(0)
    Q = h3d / (torch.sqrt(marginal_xz) * torch.sqrt(marginal_yz))
    return (Q**2).sum(dim=[0, 1]) - 1.0



"""Functions used in either generating data, or pre-processing raw data."""

import numpy as np


def sigmoid(x: np.ndarray) -> np.ndarray:
    """As there is no np.sigmoid."""
    return np.where(x >= 0, 1 / (1 + np.exp(-x)), np.exp(x) / (np.exp(x) + 1))



"""Accuracy and related metrics."""
from abc import ABC
from dataclasses import dataclass
from typing import Callable, ClassVar, Tuple

import pandas as pd
from ranzen import implements
from sklearn.metrics import accuracy_score, f1_score

from ethicml.utility import EvalTuple, Prediction

from .metric import MetricStaticName

__all__ = ["Accuracy", "F1", "SklearnMetric", "RobustAccuracy"]


class SklearnMetric(MetricStaticName, ABC):
    """Wrapper around an sklearn metric."""

    # we have to store the callable in a 1-element tuple because otherwise mypy gets confused
    sklearn_metric: ClassVar[Tuple[Callable[[pd.Series, pd.Series], float]]]

    @implements(MetricStaticName)
    def score(self, prediction: Prediction, actual: EvalTuple) -> float:
        return self.sklearn_metric[0](actual.y, prediction.hard)


@dataclass
class Accuracy(SklearnMetric):
    """Classification accuracy."""

    sklearn_metric = (accuracy_score,)
    _name: ClassVar[str] = "Accuracy"


@dataclass
class F1(SklearnMetric):
    """F1 score: harmonic mean of precision and recall."""

    sklearn_metric = (f1_score,)
    _name: ClassVar[str] = "F1"


@dataclass
class RobustAccuracy(SklearnMetric):
    """Minimum Classification accuracy across S-groups."""

    sklearn_metric = (accuracy_score,)
    apply_per_sensitive: ClassVar[bool] = False
    _name: ClassVar[str] = "Robust Accuracy"

    @implements(SklearnMetric)
    def score(self, prediction: Prediction, actual: EvalTuple) -> float:
        score_func = super().score
        return min(
            [
                score_func(prediction.get_s_subset(actual.s, _s), actual.get_s_subset(_s))
                for _s in actual.s.unique()
            ]
        )



"""Common variables / constants that make things run smoother."""
import os
from importlib import util
from pathlib import Path

__all__ = ["TORCH_AVAILABLE", "ROOT_DIR", "ROOT_PATH"]

TORCH_AVAILABLE = util.find_spec("torch") is not None

ROOT_DIR: str = os.path.abspath(os.path.join(os.path.abspath(__file__), os.pardir))
ROOT_PATH: Path = Path(__file__).parent.resolve()



"""Functions that are common to PyTorch models."""
import random
from typing import Tuple
from typing_extensions import Literal

import numpy as np
import pandas as pd

from ethicml.utility import DataTuple, TestTuple

try:
    import torch
    from torch import Tensor, nn
    from torch.utils.data import Dataset, TensorDataset
except ImportError as e:
    raise RuntimeError(
        "In order to use PyTorch, please install it following the instructions as https://pytorch.org/ . "
    ) from e


def _get_info(data: TestTuple) -> Tuple[np.ndarray, np.ndarray, int, int, pd.Index, str]:
    features = data.x.to_numpy(dtype=np.float32)
    sens_labels = data.s.to_numpy(dtype=np.float32)  # type: ignore[type-var]
    num = data.s.shape[0]
    xdim = data.x.shape[1]
    x_names = data.x.columns
    s_name = str(data.s.name)
    return features, sens_labels, num, xdim, x_names, s_name


class TestDataset(Dataset):
    """Shared Dataset for pytorch models without labels."""

    def __init__(self, data: TestTuple):
        super().__init__()
        self.x, self.s, self.num, self.xdim, self.x_names, self.s_names = _get_info(data)
        self.sdim = 1

    def __getitem__(self, index: int) -> Tuple[np.ndarray, np.ndarray]:
        """Implement __getitem__ magic method."""
        return self.x[index, ...], self.s[index, ...]

    def __len__(self) -> int:
        """Implement __len__ magic method."""
        return self.num

    @property
    def names(self) -> Tuple[pd.Index, str]:
        """Get tuple of x names and s names."""
        return self.x_names, self.s_names


class CustomDataset(Dataset):
    """Shared Dataset for pytorch models."""

    def __init__(self, data: DataTuple):
        super().__init__()
        test = data.remove_y()
        self.x, self.s, self.num, self.xdim, self.x_names, self.s_names = _get_info(test)
        self.sdim = 1
        self.y = data.y.to_numpy(dtype=np.float32)  # type: ignore[type-var]
        self.ydim = data.y.nunique()
        self.y_names = str(data.y.name)

    def __getitem__(self, index: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Implement __getitem__ magic method."""
        return self.x[index, ...], self.s[index, ...], self.y[index, ...]

    def __len__(self) -> int:
        """Implement __len__ magic method."""
        return self.num

    @property
    def names(self) -> Tuple[pd.Index, str, str]:
        """Get tuple of x names, s names and y names."""
        return self.x_names, self.s_names, self.y_names


def make_dataset_and_loader(
    data: DataTuple, *, batch_size: int, shuffle: bool, seed: int, drop_last: bool
) -> Tuple[CustomDataset, torch.utils.data.DataLoader]:
    """Given a datatuple, create a dataset and a corresponding dataloader.

    :param data:
    :param flags:
    :returns: Tuple of a pytorch dataset and dataloader.
    """

    def seed_worker(worker_id):
        worker_seed = torch.initial_seed() % 2**32
        np.random.seed(worker_seed)
        random.seed(worker_seed)

    g = torch.Generator()
    g.manual_seed(seed)

    dataset = CustomDataset(data)

    dataloader = torch.utils.data.DataLoader(
        dataset=dataset,
        batch_size=batch_size,
        shuffle=shuffle,
        drop_last=drop_last,
        worker_init_fn=seed_worker,
        generator=g,
    )
    return dataset, dataloader


def quadratic_time_mmd(x: Tensor, y: Tensor, sigma: float) -> Tensor:
    """Calculate MMD betweer 2 tensors of equal size.

    :param x: Sample 1.
    :param y: Sample 2.
    :param sigma: Scale of the RBF kernel.
    :returns: Tensor of MMD in each dim.
    """
    xx_gm = x @ x.t()
    xy_gm = x @ y.t()
    yy_gm = y @ y.t()
    x_sqnorms = torch.diagonal(xx_gm)
    y_sqnorms = torch.diagonal(yy_gm)

    def pad_first(x: Tensor) -> Tensor:
        return torch.unsqueeze(x, 0)

    def pad_second(x: Tensor) -> Tensor:
        return torch.unsqueeze(x, 1)

    gamma = 1 / (2 * sigma**2)
    # use the second binomial formula
    kernel_xx = torch.exp(-gamma * (-2 * xx_gm + pad_second(x_sqnorms) + pad_first(x_sqnorms)))
    kernel_xy = torch.exp(-gamma * (-2 * xy_gm + pad_second(x_sqnorms) + pad_first(y_sqnorms)))
    kernel_yy = torch.exp(-gamma * (-2 * yy_gm + pad_second(y_sqnorms) + pad_first(y_sqnorms)))

    xx_num = float(kernel_xx.shape[0])
    yy_num = float(kernel_yy.shape[0])

    mmd2 = (
        kernel_xx.sum() / (xx_num * xx_num)
        + kernel_yy.sum() / (yy_num * yy_num)
        - 2 * kernel_xy.sum() / (xx_num * yy_num)
    )
    return mmd2


def compute_projection_gradients(
    model: nn.Module, loss_p: Tensor, loss_a: Tensor, alpha: float
) -> None:
    """Compute the adversarial gradient projection term.

    :param model: Model whose parameters the gradients are to be computed w.r.t.
    :param loss_p: Prediction loss.
    :param loss_a: Adversarial loss.
    :param alpha: Pre-factor for adversarial loss.
    """
    grad_p = torch.autograd.grad(loss_p, model.parameters(), retain_graph=True)  # type: ignore[arg-type]
    grad_a = torch.autograd.grad(loss_a, model.parameters(), retain_graph=True)  # type: ignore[arg-type]

    def _proj(a: Tensor, b: Tensor) -> Tensor:
        return b * torch.sum(a * b) / torch.sum(b * b)

    grad_p = tuple(p - _proj(p, a) - alpha * a for p, a in zip(grad_p, grad_a))

    for param, grad in zip(model.parameters(), grad_p):
        param.grad = grad


class PandasDataSet(TensorDataset):
    """Pandas Dataset."""

    def __init__(self, *dataframes):
        tensors = (self._df_to_tensor(df) for df in dataframes)
        super().__init__(*tensors)

    def _df_to_tensor(self, df):
        if isinstance(df, pd.Series):
            df = df.to_frame('dummy')
        return torch.from_numpy(df.values).float()


class LinearModel(torch.nn.Module):
    """Define linear model."""

    def __init__(self, in_shape: int = 1, out_shape: int = 2):
        super().__init__()
        self.in_shape = in_shape
        self.out_shape = out_shape

        self.base_model = nn.Sequential(
            nn.Linear(self.in_shape, self.out_shape, bias=True),
        )

    def build_model(self) -> None:
        """Build Model."""

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward."""
        return torch.squeeze(self.base_model(x))


class DeepModel(torch.nn.Module):
    """Define deep neural net model for classification."""

    def __init__(self, in_shape: int = 1, out_shape: int = 1):
        super().__init__()
        self.in_shape = in_shape
        self.dim_h = 64
        self.dropout = 0.5
        self.out_shape = out_shape
        self.build_model()

    def build_model(self) -> None:
        """Build Model."""
        self.base_model = nn.Sequential(
            nn.Linear(self.in_shape, self.dim_h, bias=True),
            nn.ReLU(),
            nn.Dropout(self.dropout),
            nn.Linear(self.dim_h, self.out_shape, bias=True),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward."""
        return torch.squeeze(self.base_model(x))


class DeepRegModel(torch.nn.Module):
    """Define deep model for regression."""

    def __init__(self, in_shape: int = 1, out_shape: int = 1):
        super().__init__()
        self.in_shape = in_shape
        self.dim_h = 64  # in_shape*10
        self.out_shape = out_shape
        self.build_model()

    def build_model(self) -> None:
        """Build model."""
        self.base_model = nn.Sequential(
            nn.Linear(self.in_shape, self.dim_h, bias=True),
            nn.ReLU(),
            nn.Linear(self.dim_h, self.out_shape, bias=True),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward."""
        return torch.squeeze(self.base_model(x))


class DeepProbaModel(torch.nn.Module):
    """Define deep regression model, used by the fair dummies test."""

    def __init__(self, in_shape: int = 1):
        super().__init__()
        self.in_shape = in_shape
        self.dim_h = 64  # in_shape*10
        self.dropout = 0.5
        self.out_shape = 1
        self.build_model()

    def build_model(self) -> None:
        """Build Model."""
        self.base_model = nn.Sequential(
            nn.Linear(self.in_shape, self.dim_h, bias=True),
            nn.ReLU(),
            nn.Dropout(self.dropout),
            nn.Linear(self.dim_h, self.out_shape, bias=True),
            nn.Sigmoid(),
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward."""
        return torch.squeeze(self.base_model(x))


class GeneralLearner:
    """General Learner."""

    def __init__(
        self,
        lr: float,
        epochs: int,
        cost_func: nn.Module,
        in_shape: int,
        batch_size: int,
        model_type: Literal["deep_proba", "deep_regression"],
        out_shape: int = 1,
    ):

        # input dim
        self.in_shape = in_shape

        # output dim
        self.out_shape = out_shape

        # learning rate
        self.lr = lr

        # number of epochs
        self.epochs = epochs

        # cost to minimize
        self.cost_func = cost_func

        self.rng = np.random.default_rng(0)

        # define a predictive model
        self.model_type = model_type
        if self.model_type == "deep_proba":
            self.model: nn.Module = DeepProbaModel(in_shape=in_shape)
        elif self.model_type == "deep_regression":
            self.model = DeepModel(in_shape=in_shape, out_shape=self.out_shape)
        else:
            raise NotImplementedError

        # optimizer
        self.optimizer = torch.optim.SGD(self.model.parameters(), lr=self.lr, momentum=0.9)

        # minibatch size
        self.batch_size = batch_size

    def internal_epoch(self, dataloader: torch.utils.data.DataLoader) -> np.ndarray:
        """Fit a model by sweeping over all data points."""
        # fit pred func
        epoch_losses = []
        for x, s, y in dataloader:
            self.optimizer.zero_grad()
            # utility loss
            batch_yhat = self.model(x)
            loss = self.cost_func(batch_yhat, y.long())
            loss.backward()
            self.optimizer.step()
            epoch_losses.append(loss.cpu().detach().numpy())
        return np.mean(epoch_losses)

    def run_epochs(self, dataloader: torch.utils.data.DataLoader) -> None:
        """Run epochs."""
        for _ in range(self.epochs):
            self.internal_epoch(dataloader)

    def fit(self, train: DataTuple, seed: int) -> None:
        """Fit a model on training data."""
        self.model.train()
        _, train_loader = make_dataset_and_loader(
            train, batch_size=self.batch_size, shuffle=True, seed=seed, drop_last=True
        )
        self.run_epochs(train_loader)

    @torch.no_grad()
    def predict(self, x: torch.Tensor) -> torch.Tensor:
        """Predict output."""
        self.model.eval()
        xp = torch.from_numpy(x).float()
        yhat = self.model(xp)
        return yhat.detach().numpy()



"""Wrapper for SKLearn implementation of MLP."""
from dataclasses import dataclass, field
from typing import ClassVar, Tuple

import numpy as np
from ranzen import implements
from sklearn.neural_network import MLPClassifier

from ethicml.algorithms.inprocess.in_algorithm import InAlgorithmDC
from ethicml.utility import DataTuple, Prediction, SoftPrediction, TestTuple

__all__ = ["MLP"]


@dataclass
class MLP(InAlgorithmDC):
    """Multi-layer Perceptron.

    This is a wraper around the SKLearn implementation of the MLP.
    Documentation: https://scikit-learn.org/stable/modules/generated/sklearn.neural_network.MLPClassifier.html

    :param hidden_layer_sizes: The number of neurons in each hidden layer.
    """

    is_fairness_algo: ClassVar[bool] = False
    hidden_layer_sizes: Tuple[int, ...] = field(
        default_factory=lambda: MLPClassifier().hidden_layer_sizes
    )
    batch_size: int = 32
    lr: float = 1e-3

    @implements(InAlgorithmDC)
    def get_name(self) -> str:
        return "MLP"

    @implements(InAlgorithmDC)
    def fit(self, train: DataTuple, seed: int = 888) -> InAlgorithmDC:
        self.clf = select_mlp(
            self.hidden_layer_sizes, seed=seed, lr=self.lr, batch_size=self.batch_size
        )
        self.clf.fit(train.x, train.y.to_numpy().ravel())
        return self

    @implements(InAlgorithmDC)
    def predict(self, test: TestTuple) -> Prediction:
        return SoftPrediction(soft=self.clf.predict_proba(test.x), info=self.get_hyperparameters())

    @implements(InAlgorithmDC)
    def run(self, train: DataTuple, test: TestTuple, seed: int = 888) -> Prediction:
        clf = select_mlp(self.hidden_layer_sizes, seed=seed, lr=self.lr, batch_size=self.batch_size)
        clf.fit(train.x, train.y.to_numpy().ravel())
        return SoftPrediction(soft=clf.predict_proba(test.x), info=self.get_hyperparameters())


def select_mlp(
    hidden_layer_sizes: Tuple[int, ...], seed: int, lr: float, batch_size: int
) -> MLPClassifier:
    """Create MLP model for the given parameters.

    :param hidden_layer_sizes: The number of neurons in each hidden layer.
    :param seed: The seed for the random number generator.
    """
    random_state = np.random.RandomState(seed=seed)
    return MLPClassifier(
        hidden_layer_sizes=hidden_layer_sizes,
        solver='adam',
        random_state=random_state,
        learning_rate_init=lr,
        batch_size=batch_size,
    )



"""Module for metrics which can be applied to prediction results.

Some example code

.. code:: python

   import ethicml as em

   em.run_metrics(predictions, test_data, metrics=[em.Accuracy(), em.TPR(), em.ProbPos()])
"""
from .accuracy import *
from .anti_spur import *
from .average_odds import *
from .balanced_accuracy import *
from .bcr import *
from .confusion_matrix import *
from .cv import *
from .dependence_measures import *
from .fnr import *
from .fpr import *
from .hsic import *
from .metric import *
from .npv import *
from .per_sensitive_attribute import *
from .ppv import *
from .prob_neg import *
from .prob_outcome import *
from .prob_pos import *
from .theil import *
from .tnr import *
from .tpr import *



"""Generate Lipton's Synthetic Data."""
from pathlib import Path

import numpy as np
import pandas as pd

SEED = 8
SAMPLES = 2_000


def main() -> None:
    """Generate the data as specified in section 4.1 of https://arxiv.org/pdf/1711.07076.pdf ."""
    rng = np.random.default_rng(SEED)

    z = rng.binomial(1, 0.5, SAMPLES)

    hair_length_0: np.ndarray = 35 * rng.beta(2, 7, SAMPLES)
    hair_length_1: np.ndarray = 35 * rng.beta(2, 2, SAMPLES)
    hair_length = np.where(z == 1, hair_length_1, hair_length_0)

    pois = rng.poisson((25 + 6 * z))
    norm = rng.normal(20, 0.2, SAMPLES)
    work_experience: np.ndarray = pois - norm

    p = 1 / (1 + (np.exp(-(-25.5 + 2.5 * work_experience))))

    y = 2 * rng.binomial(1, p) - 1

    df = pd.DataFrame(
        data={
            "hair_length": hair_length,
            "work_experience": work_experience,
            "sens": z,
            "hired": y,
        }
    )

    df.to_csv(str(Path(__file__).parent / "lipton.csv"), index=False)


if __name__ == '__main__':
    main()



"""For assessing ProbNeg."""
from dataclasses import dataclass
from typing import ClassVar

from ranzen import implements

from ethicml.utility import EvalTuple, Prediction

from .confusion_matrix import CfmMetric
from .metric import Metric

__all__ = ["ProbNeg"]


@dataclass
class ProbNeg(CfmMetric):
    """Probability of negative prediction."""

    _name: ClassVar[str] = "prob_neg"

    @implements(Metric)
    def score(self, prediction: Prediction, actual: EvalTuple) -> float:
        t_neg, _, f_neg, _ = self.confusion_matrix(prediction=prediction, actual=actual)
        return (t_neg + f_neg) / prediction.hard.size



"""Data structure for all datasets that come with the framework."""
import typing
from abc import abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Tuple, Union
from typing_extensions import Literal, TypedDict

import pandas as pd
from ranzen import implements

from ethicml.common import ROOT_PATH
from ethicml.utility import DataTuple, undo_one_hot

from .util import (
    LabelSpec,
    filter_features_by_prefixes,
    get_discrete_features,
    label_spec_to_feature_list,
)

__all__ = ["Dataset", "FeatureSplit", "LoadableDataset"]


class FeatureSplit(TypedDict):
    """A dictionary of the list of columns that belong to the feature groups."""

    x: List[str]
    s: List[str]
    y: List[str]


class Dataset:
    """Data structure that holds all the information needed to load a given dataset.

    :param discard_non_one_hot: If some entries in s or y are not correctly one-hot encoded, discard
        those.
    :param map_to_binary: If True, convert labels from {-1, 1} to {0, 1}.
    """

    def __init__(
        self,
        name: str,
        filename_or_path: Union[str, Path],
        features: Sequence[str],
        cont_features: Sequence[str],
        sens_attr_spec: Union[str, LabelSpec],
        class_label_spec: Union[str, LabelSpec],
        discrete_only: bool,
        num_samples: int,
        discard_non_one_hot: bool = False,
        map_to_binary: bool = False,
        s_prefix: Optional[Sequence[str]] = None,
        class_label_prefix: Optional[Sequence[str]] = None,
        discrete_feature_groups: Optional[Dict[str, List[str]]] = None,
    ) -> None:
        self._name = name
        self._features = features
        self._sens_attr_spec = sens_attr_spec
        self._class_label_spec = class_label_spec
        self._num_samples = num_samples
        self._discrete_only = discrete_only
        self._s_prefix = [] if s_prefix is None else s_prefix
        self._class_label_prefix = [] if class_label_prefix is None else class_label_prefix
        self._discard_non_one_hot = discard_non_one_hot
        self._map_to_binary = map_to_binary
        self._discrete_feature_groups = discrete_feature_groups
        self._raw_file_name_or_path = filename_or_path
        self._cont_features_unfiltered = cont_features

    @property
    def name(self) -> str:
        """Name of the dataset."""
        return self._name

    @property
    def sens_attrs(self) -> List[str]:
        """Get the list of sensitive attributes."""
        if isinstance(self._sens_attr_spec, str):
            return [self._sens_attr_spec]
        assert isinstance(self._sens_attr_spec, dict)
        return label_spec_to_feature_list(self._sens_attr_spec)

    @property
    def class_labels(self) -> List[str]:
        """Get the list of class labels."""
        if isinstance(self._class_label_spec, str):
            return [self._class_label_spec]
        assert isinstance(self._class_label_spec, dict)
        return label_spec_to_feature_list(self._class_label_spec)

    @property
    def filepath(self) -> Path:
        """Filepath from which to load the data."""
        if isinstance(self._raw_file_name_or_path, Path):
            return self._raw_file_name_or_path
        else:
            return ROOT_PATH / "data" / "csvs" / self._raw_file_name_or_path

    @property
    def features_to_remove(self) -> List[str]:
        """Features that have to be removed from x."""
        to_remove: List[str] = []
        to_remove += self._s_prefix
        to_remove += self._class_label_prefix
        if self._discrete_only:
            to_remove += self.continuous_features
        return to_remove

    @property
    def ordered_features(self) -> FeatureSplit:
        """Return an order features dictionary.

        This should have separate entries for the features, the labels and the
        sensitive attributes, but the x features are ordered so first are the discrete features,
        then the continuous.
        """
        ordered = self.discrete_features + self.continuous_features
        return {
            "x": filter_features_by_prefixes(ordered, self.features_to_remove),
            "s": self.sens_attrs,
            "y": self.class_labels,
        }

    @property
    def feature_split(self) -> FeatureSplit:
        """Return a feature split dictionary.

        This should have separate entries for the features, the labels and the sensitive attributes.
        """
        features_to_remove = self.features_to_remove

        return {
            "x": filter_features_by_prefixes(self._features, features_to_remove),
            "s": self.sens_attrs,
            "y": self.class_labels,
        }

    @property
    def continuous_features(self) -> List[str]:
        """List of features that are continuous."""
        return filter_features_by_prefixes(self._cont_features_unfiltered, self.features_to_remove)

    @property
    def discrete_features(self) -> List[str]:
        """List of features that are discrete."""
        return get_discrete_features(
            list(self._features), self.features_to_remove, self.continuous_features
        )

    @property
    def disc_feature_groups(self) -> Optional[Dict[str, List[str]]]:
        """Return Dictionary of feature groups."""
        if self._discrete_feature_groups is None:
            return None
        dfgs = self._discrete_feature_groups
        return {k: v for k, v in dfgs.items() if k not in self.features_to_remove}

    def __len__(self) -> int:
        """Return number of elements in the dataset."""
        return self._num_samples

    @abstractmethod
    def load(self, ordered: bool = False, labels_as_features: bool = False) -> DataTuple:
        """Load the dataset.

        :param ordered:  (Default: False)
        :param labels_as_features:  (Default: False)
        """
        raise NotImplementedError("Dataset is abstract.")

    def _one_hot_encode_and_combine(
        self, attributes: pd.DataFrame, label_type: Literal["s", "y"]
    ) -> Tuple[pd.Series, Optional[pd.Series]]:
        """Construct a new label according to the LabelSpecs.

        :param attributes: DataFrame containing the attributes.
        :param label_type: Type of label to construct.
        """
        mask = None  # the mask is needed when we have to discard samples

        label_mapping = self._sens_attr_spec if label_type == "s" else self._class_label_spec
        if isinstance(label_mapping, str):
            return attributes[label_mapping], mask

        # create a Series of zeroes with the same length as the dataframe
        combination: pd.Series = pd.Series(  # type: ignore[call-overload]
            0, index=range(len(attributes)), name=",".join(label_mapping)
        )

        for name, spec in label_mapping.items():
            if len(spec.columns) > 1:  # data is one-hot encoded
                raw_values = attributes[spec.columns]
                if self._discard_non_one_hot:
                    # only use those samples where exactly one of the specified attributes is true
                    mask = raw_values.sum(axis="columns") == 1
                else:
                    assert (raw_values.sum(axis="columns") == 1).all(), f"{name} is not one-hot"
                values = undo_one_hot(raw_values)
            else:
                values = attributes[spec.columns[0]]
            combination += spec.multiplier * values
        return combination, mask

    @staticmethod
    def _from_dummies(
        data: pd.DataFrame, categorical_cols: Dict[str, Sequence[str]]
    ) -> pd.DataFrame:
        out = data.copy()

        for col_parent, filter_col in categorical_cols.items():
            if len(filter_col) > 1:
                undummified = (
                    out[filter_col].idxmax(axis=1).apply(lambda x: x.split(f"{col_parent}_", 1)[1])
                )

                out[col_parent] = undummified
                out.drop(filter_col, axis=1, inplace=True)

        return out

    def expand_labels(self, label: pd.Series, label_type: Literal["s", "y"]) -> pd.DataFrame:
        """Expand a label in the form of an index into all the subfeatures.

        :param label: DataFrame containing the labels.
        :param label_type: Type of label to expand.
        """
        label_mapping = self._sens_attr_spec if label_type == "s" else self._class_label_spec
        assert not isinstance(label_mapping, str)
        label_mapping: LabelSpec  # redefine the variable for mypy's sake

        # first order the multipliers; this is important for disentangling the values
        names_ordered = sorted(label_mapping, key=lambda name: label_mapping[name].multiplier)

        final_df = {}
        for i, name in enumerate(names_ordered):
            spec = label_mapping[name]
            value = label
            if i + 1 < len(names_ordered):
                next_group = label_mapping[names_ordered[i + 1]]
                value = label % next_group.multiplier
            value = value // spec.multiplier
            value.replace(list(range(len(spec.columns))), spec.columns, inplace=True)
            restored = pd.get_dummies(value)
            final_df[name] = restored  # for the multi-level column index

        return pd.concat(final_df, axis=1)


@dataclass(init=False)
class LoadableDataset(Dataset):
    """Dataset that uses the default load function."""

    # TODO: actually move the loading code into this class. This will require some wider changes.
    discrete_only: bool = False
    invert_s: bool = False

    @implements(Dataset)
    def load(self, ordered: bool = False, labels_as_features: bool = False) -> DataTuple:
        """Load dataset from its CSV file.

        :param ordered: if True, return features such that discrete come first, then continuous
        :param labels_as_features: if True, the s and y labels are included in the x features
        :returns: DataTuple with dataframes of features, labels and sensitive attributes
        """
        dataframe: pd.DataFrame = pd.read_csv(self.filepath, nrows=self._num_samples)
        assert isinstance(dataframe, pd.DataFrame)

        feature_split = self.ordered_features if ordered else self.feature_split
        if labels_as_features:
            feature_split_x = feature_split["x"] + feature_split["s"] + feature_split["y"]
        else:
            feature_split_x = feature_split["x"]

        # =========================================================================================
        # Check whether we have to generate some complementary columns for binary features.
        # This happens when we have for example several races: race-asian-pac-islander etc, but we
        # want to have a an attribute called "race_other" that summarizes them all. Now the problem
        # is that this cannot be done before this point, because only here have we actually loaded
        # the data. So, we have to do it here, with all the information we can piece together.

        disc_feature_groups = self._discrete_feature_groups
        if disc_feature_groups is not None:
            for group in disc_feature_groups.values():
                if len(group) == 1:
                    continue
                for feature in group:
                    if feature in dataframe.columns:
                        continue  # nothing to do
                    missing_feature = feature

                    existing_features = [other for other in group if other in dataframe.columns]
                    assert len(existing_features) == len(group) - 1, "at most 1 feature missing"
                    # the dummy feature is the inverse of the existing feature
                    or_combination = dataframe[existing_features[0]] == 1
                    for other in existing_features[1:]:
                        or_combination |= dataframe[other] == 1
                    inverse: pd.Series = 1 - or_combination
                    dataframe = pd.concat(
                        [dataframe, inverse.to_frame(name=missing_feature)], axis="columns"
                    )

        # =========================================================================================
        x_data = dataframe[feature_split_x]
        s_df = dataframe[feature_split["s"]]
        y_df = dataframe[feature_split["y"]]

        if self._map_to_binary:
            s_df = (s_df + 1) // 2  # map from {-1, 1} to {0, 1}
            y_df = (y_df + 1) // 2  # map from {-1, 1} to {0, 1}

        if self.invert_s:
            assert s_df.nunique().values[0] == 2, "s must be binary"
            s_df = 1 - s_df

        # the following operations remove rows if a label group is not properly one-hot encoded
        s_data, s_mask = self._one_hot_encode_and_combine(s_df, label_type="s")
        if s_mask is not None:
            x_data = x_data.loc[s_mask].reset_index(drop=True)
            s_data = s_data.loc[s_mask].reset_index(drop=True)
            y_df = y_df.loc[s_mask].reset_index(drop=True)
        y_data, y_mask = self._one_hot_encode_and_combine(y_df, label_type="y")
        if y_mask is not None:
            x_data = x_data.loc[y_mask].reset_index(drop=True)
            s_data = s_data.loc[y_mask].reset_index(drop=True)
            y_data = y_data.loc[y_mask].reset_index(drop=True)

        return DataTuple.from_df(x=x_data, s=s_data, y=y_data, name=self.name)

    @typing.no_type_check
    def load_aif(self):  # Returns aif.360 Standard Dataset
        """Load the dataset as an AIF360 dataset.

        Experimental.
        Requires the aif360 library.

        Ignores the type check as the return type is not yet defined.

        """
        from aif360.datasets import StandardDataset

        data = self.load()
        if self.disc_feature_groups is not None:
            data_collapsed = self._from_dummies(data.x, self.disc_feature_groups)
            data = data.replace(x=data_collapsed)
        df = data.data

        return StandardDataset(
            df=df,
            label_name=data.y_column,
            favorable_classes=lambda x: x > 0,
            protected_attribute_names=[data.s_column],
            privileged_classes=[lambda x: x == 1],
            categorical_features=self.disc_feature_groups.keys()
            if self.disc_feature_groups is not None
            else None,
        )



"""For assessing NPV."""
from dataclasses import dataclass
from typing import ClassVar

from ranzen import implements

from ethicml.utility import EvalTuple, Prediction

from .confusion_matrix import CfmMetric
from .metric import Metric

__all__ = ["NPV"]


@dataclass
class NPV(CfmMetric):
    """Negative predictive value."""

    _name: ClassVar[str] = "NPV"

    @implements(Metric)
    def score(self, prediction: Prediction, actual: EvalTuple) -> float:
        t_neg, _, f_neg, _ = self.confusion_matrix(prediction=prediction, actual=actual)
        return t_neg / (t_neg + f_neg)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ethicml',
 'ethicml.algorithms',
 'ethicml.algorithms.inprocess',
 'ethicml.algorithms.postprocess',
 'ethicml.algorithms.preprocess',
 'ethicml.data',
 'ethicml.data.csvs',
 'ethicml.data.tabular_data',
 'ethicml.evaluators',
 'ethicml.implementations',
 'ethicml.implementations.adv_debiasing_modules',
 'ethicml.implementations.dro_modules',
 'ethicml.implementations.fair_dummies_modules',
 'ethicml.implementations.hgr_modules',
 'ethicml.implementations.hgr_modules.facl',
 'ethicml.implementations.vfae_modules',
 'ethicml.metrics',
 'ethicml.preprocessing',
 'ethicml.utility',
 'ethicml.visualisation']

package_data = \
{'': ['*']}

install_requires = \
['GitPython>=3.1.20,<4.0.0',
 'Pillow>=9.0.1,<10.0.0',
 'folktables>=0.0.11,<0.0.12',
 'gitdb2==4.0.2',
 'joblib>=1.1.0,<2.0.0',
 'matplotlib>=3.0.2',
 'numpy>=1.20.1',
 'pandas>=1.0,<2.0',
 'pipenv>=2018.11.26',
 'ranzen>=1.1.1,<2.0.0',
 'scikit-learn>=0.20.1',
 'scipy>=1.7.2,<2.0.0',
 'seaborn>=0.9.0',
 'smmap2==3.0.1',
 'teext>=0.1.3,<0.2.0',
 'tqdm>=4.31.1',
 'typing-extensions>=4.0']

extras_require = \
{'all': ['fairlearn==0.7.0', 'cloudpickle>=2.0.0,<3.0.0'],
 'ci': ['fairlearn==0.7.0',
        'cloudpickle>=2.0.0,<3.0.0',
        'pytest>=6.2.2,<8.0.0',
        'pytest-cov>=2.6,<4.0',
        'omegaconf>=2.2.1']}

setup_kwargs = {
    'name': 'ethicml',
    'version': '1.0.0a0',
    'description': "EthicML is a library for performing and assessing algorithmic fairness. Unlike other libraries, EthicML isn't an education tool, but rather a researcher's toolkit.",
    'long_description': '# EthicML: A featureful framework for developing fair algorithms\n\n[![Checked with mypy](http://www.mypy-lang.org/static/mypy_badge.svg)](http://mypy-lang.org/)\n![](https://github.com/predictive-analytics-lab/EthicML/workflows/EthicML%20CI/badge.svg)\n\n\n\nEthicML is a library for performing and assessing __algorithmic fairness__.\nUnlike other libraries, EthicML isn\'t an education tool, but rather a __researcher__\'s toolkit.\n\nOther algorthimic fairness packages are useful, but given that we primarily do research,\na lot of the work we do doesn\'t fit into some nice box.\nFor example, we might want to use a \'fair\' pre-processing method on the data before training a classifier on it.\nWe may still be experimenting and only want part of the framework to execute,\nor we may want to do hyper-parameter optimization.\nWhilst other frameworks can be modified to do these tasks,\nyou end up with hacked-together approaches that don\'t lend themselves to be built on in the future.\nBecause of this, we built EthicML, a fairness toolkit for researchers.\n\nFeatures include:\n- Support for multiple sensitive attributes\n- Vision datasets\n- Codebase typed with mypy\n- Tested code\n- Reproducible results\n\n### Why not use XXX?\n\nThere are an increasing number of other options,\nIBM\'s fair-360, Aequitas, EthicalML/XAI, Fairness-Comparison and others.\nThey\'re all great at what they do, they\'re just not right for us.\nWe will however be influenced by them.\nWhere appropriate, we even subsume some of these libraries.\n\n## Installation\n\nEthicML requires Python >= 3.7.\nTo install EthicML, just do\n```\npip3 install ethicml\n```\n\nIf you want to use the method by Agarwal et al., you have to explicitly install _all_ dependencies:\n```\npip3 install \'ethicml[all]\'\n```\n(The quotes are needed in `zsh` and will also work in `bash`.)\n\n**Attention**: In order to use all features of EthicML, PyTorch needs to be installed separately.\nWe are not including PyTorch as a requirement of EthicML,\nbecause there are many different versions for different systems.\n\n## Documentation\n\nThe documentation can be found here: https://wearepal.ai/EthicML/\n\n## Design Principles\n\n```mermaid\nflowchart LR\n    A(Datasets) -- load --> B(Data tuples);\n    B --> C[evaluate_models];\n    G(Algorithms) --> C;\n    C --> D(Metrics);\n```\n\nKeep things simple.\n\n### The Triplet\n\nGiven that we\'re considering fairness, the base of the toolbox is the triplet {x, s, y}\n\n- X - Features\n- S - Sensitive Label\n- Y - Class Label\n\n__Developer note__: All methods must assume S and Y are multi-class.\n\nWe use a DataTuple class to contain the triplet\n\n```python\ntriplet = DataTuple(x: pandas.DataFrame, s: pandas.DataFrame, y: pandas.DataFrame)\n```\n\nIn addition, we have a variation: the TestTuple which contains the pair\n```python\npair = TestTuple(x: pandas.DataFrame, s: pandas.DataFrame)\n```\nThis is to reduce the risk of a user accidentally evaluating performance on their training set.\n\nUsing dataframes may be a little inefficient,\nbut given the amount of splicing on conditions that we\'re doing, it feels worth it.\n\n### Separation of Methods\n\nWe purposefully keep pre, during and post algorithm methods separate. This is because they have different return types.\n\n```python\npre_algorithm.run(train: DataTuple, test: TestTuple)  # -> Tuple[DataTuple, TestTuple]\nin_algorithm.run(train: DataTuple, test: TestTuple)  # -> Prediction\npost_algorithm.run(train_prediction: Prediction, train: DataTuple, test_prediction: Prediction, test: TestTuple)  # -> Prediction\n```\nwhere `Prediction` holds a pandas.Series of the class label.\nIn the case of a "soft" output, `SoftPrediction` extends `Prediction` and provides a mapping from\n"soft" to "hard" labels.\nSee the documentation for more details.\n\n### General Rules of Thumb\n\n- Mutable data structures are bad.\n- At the very least, functions should be Typed.\n- Readability > Efficiency.\n- Warnings must be addressed.\n- Always write tests first.\n\n## Future Plans\n\nThe aim is to make EthicML operate on 2 levels.\n\n1. We want a high-level API so that a user can define a new model or metric, then get publication-ready\nresults in just a couple of lines of code.\n2. We understand that truly ground-breaking work sometimes involves tearing up the rulebook.\nTherefore, we want to also expose a lower-level API so that a user can make use of as much, or little of the library\nas is suitable for them.\n\nWe\'ve built everything with this philosophy in mind, but acknowledge that we still have a way to go.\n\n# Contributing\n\nIf you\'re interest in this research area, we\'d love to have you aboard.\nFor more details check out [CONTRIBUTING.md](./CONTRIBUTING.md).\nWhether your skills are in coding-up papers you\'ve read, writing tutorials, or designing a logo, please reach out.\n\n## Development\nInstall development dependencies with `pip install -e .[dev]`\n\nTo use the pre-commit hooks run `pre-commit install`\n',
    'author': 'PAL',
    'author_email': 'info@wearepal.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/predictive-analytics-lab/EthicML',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7.1,<3.11',
}


setup(**setup_kwargs)



"""Wrapper for calling Kamishima model."""
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import NamedTuple, Optional, Union

import numpy as np
import pandas as pd
from ranzen import implements

from ethicml.algorithms.inprocess.in_algorithm import HyperParamType, InAlgorithm
from ethicml.algorithms.inprocess.installed_model import InstalledModel
from ethicml.utility import DataTuple, Prediction, TestTuple

__all__ = ["Kamishima"]


class _FitInfo(NamedTuple):
    """Information that is stored after the model has been fit."""

    min_class_label: int
    model_path: Path


class Kamishima(InstalledModel):
    """Model that calls Kamishima's code.

    Based on Algo-Fairness
    https://github.com/algofairness/fairness-comparison/blob/master/fairness/algorithms/kamishima/KamishimaAlgorithm.py

    :param eta: Tolerance.
    """

    def __init__(self, *, eta: float = 1.0):
        super().__init__(
            name="Kamishima",
            dir_name="kamishima",
            url="https://github.com/predictive-analytics-lab/kamfadm.git",
            top_dir="kamfadm",
        )
        self.eta = eta
        self._fit_info: Optional[_FitInfo] = None

    @implements(InAlgorithm)
    def get_hyperparameters(self) -> HyperParamType:
        return {"eta": self.eta}

    @implements(InAlgorithm)
    def run(self, train: DataTuple, test: TestTuple, seed: int = 888) -> Prediction:
        with TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            fit_info = self._fit(train, tmp_path, seed)
            return self._predict(test, fit_info, tmp_path)

    @implements(InAlgorithm)
    def fit(self, train: DataTuple, seed: int = 888) -> "Kamishima":
        with TemporaryDirectory() as tmpdir:
            self._fit_info = self._fit(train, Path(tmpdir), seed, model_dir=self._code_path)
        return self

    def _fit(
        self, train: DataTuple, tmp_path: Path, seed: int, model_dir: Optional[Path] = None
    ) -> _FitInfo:
        train_path = tmp_path / "train.txt"
        _create_file_in_kamishima_format(train, train_path)
        min_class_label: int = train.y.min()
        model_path = (model_dir if model_dir is not None else tmp_path) / "model"

        script = self._code_path / "train_pr.py"
        cmd = [script, "-e", self.eta, "-i", train_path, "-o", model_path, "--quiet"]
        self.call_script([str(e) for e in cmd])
        return _FitInfo(min_class_label=min_class_label, model_path=model_path)

    @implements(InAlgorithm)
    def predict(self, test: TestTuple) -> Prediction:
        assert self._fit_info is not None, "call fit() before calling predict()"
        with TemporaryDirectory() as tmpdir:
            return self._predict(test, self._fit_info, Path(tmpdir))

    def _predict(self, test: TestTuple, fit_info: _FitInfo, tmp_path: Path) -> Prediction:
        test_path = tmp_path / "test.txt"
        _create_file_in_kamishima_format(test, test_path)
        output_path = str(tmp_path / "output.txt")
        script = self._code_path / "predict_lr.py"
        cmd = [script, "-i", test_path, "-m", fit_info.model_path, "-o", output_path, "--quiet"]
        self.call_script([str(e) for e in cmd])
        output = np.loadtxt(output_path)
        predictions = output[:, 1].astype(np.float32)
        # except RuntimeError:
        #     predictions = np.ones_like(test.y.to_numpy())

        to_return = pd.Series(predictions)
        to_return = to_return.astype(int)

        if to_return.min() != to_return.max():
            to_return = to_return.replace(to_return.min(), fit_info.min_class_label)
        return Prediction(hard=to_return)


def _create_file_in_kamishima_format(data: Union[DataTuple, TestTuple], file_path: Path) -> None:
    """Create a text file with the data.

    :param data: Data to write to the file.
    :param file_path: Path to the file.
    """
    if isinstance(data, DataTuple):
        result = pd.concat([data.x, data.s, data.y], axis="columns").to_numpy().astype(np.float64)
    else:
        zeros = pd.DataFrame([0 for _ in range(data.x.shape[0])], columns=["y"])
        result = pd.concat([data.x, data.s, zeros], axis="columns").to_numpy().astype(np.float64)
    np.savetxt(file_path, result)



"""Density Estimations."""
from math import pi, sqrt

import torch


class Kde:
    """A Gaussian KDE implemented in pytorch for the gradients to flow in pytorch optimization.

    Keep in mind that KDE are not scaling well with the number of dimensions and this implementation is not really
    optimized...
    """

    def __init__(self, x_train: torch.Tensor):
        n, d = x_train.shape

        self.n = n
        self.d = d

        self.bandwidth = (n * (d + 2) / 4.0) ** (-1.0 / (d + 4))
        self.std = self.bandwidth

        self.train_x = x_train

    def pdf(self, x: torch.Tensor) -> torch.Tensor:
        """Probability Density Function."""
        s = x.shape
        d = s[-1]
        s = s[:-1]
        assert d == self.d

        data = x.unsqueeze(-2)

        train_x = _unsqueeze_multiple_times(self.train_x, 0, len(s))

        return (
            (torch.exp(-((data - train_x).norm(dim=-1) ** 2 / (self.bandwidth**2) / 2))).mean(
                dim=-1
            )
            / sqrt(2 * pi)
            / self.bandwidth
        )


def _unsqueeze_multiple_times(input: torch.Tensor, axis: int, times: int) -> torch.Tensor:
    """Utils function to unsqueeze tensor to avoid cumbersome code.

    :param input: A pytorch Tensor of dimensions (D_1,..., D_k)
    :param axis: the axis to unsqueeze repeatedly
    :param times: the number of repetitions of the unsqueeze
    :return: the unsqueezed tensor. ex: dimensions (D_1,... D_i, 0,0,0, D_{i+1}, ... D_k) for unsqueezing 3x axis i.
    """
    output = input
    for _ in range(times):
        output = output.unsqueeze(axis)
    return output



"""Module for algorithms that pre-process the data in some way."""

from .adjust_labels import *
from .biased_split import *
from .domain_adaptation import *
from .feature_binning import *
from .scaling import *
from .splits import *



"""Class to describe features of the UCI Credit dataset."""
from dataclasses import dataclass
from enum import Enum
from typing import ClassVar, Type

from ..dataset import LoadableDataset
from ..util import flatten_dict

__all__ = ["Credit", "CreditSplits"]


class CreditSplits(Enum):
    """Splits for the Credit dataset."""

    SEX = "Sex"
    CUSTOM = "Custom"


@dataclass
class Credit(LoadableDataset):
    """UCI Credit Card dataset."""

    split: CreditSplits = CreditSplits.SEX

    Splits: ClassVar[Type[CreditSplits]] = CreditSplits
    """Shorthand for the Enum that defines the splits associated with this class."""

    def __post_init__(self) -> None:
        disc_feature_groups = {
            "SEX": ["SEX"],
            "EDUCATION": [
                "EDUCATION_1",
                "EDUCATION_2",
                "EDUCATION_3",
                "EDUCATION_4",
                "EDUCATION_5",
                "EDUCATION_6",
            ],
            "MARRIAGE": [
                "MARRIAGE_1",
                "MARRIAGE_2",
                "MARRIAGE_3",
            ],
            "default-payment-next-month": ["default-payment-next-month"],
        }
        disc_features = flatten_dict(disc_feature_groups)

        continuous_features = [
            "LIMIT_BAL",
            "AGE",
            "PAY_0",
            "PAY_2",
            "PAY_3",
            "PAY_4",
            "PAY_5",
            "PAY_6",
            "BILL_AMT1",
            "BILL_AMT2",
            "BILL_AMT3",
            "BILL_AMT4",
            "BILL_AMT5",
            "BILL_AMT6",
            "PAY_AMT1",
            "PAY_AMT2",
            "PAY_AMT3",
            "PAY_AMT4",
            "PAY_AMT5",
            "PAY_AMT6",
        ]

        if self.split is CreditSplits.SEX:
            sens_attr_spec = "SEX"
            s_prefix = ["SEX"]
            class_label_spec = "default-payment-next-month"
            class_label_prefix = ["default-payment-next-month"]
        elif self.split is CreditSplits.CUSTOM:
            sens_attr_spec = ""
            s_prefix = []
            class_label_spec = ""
            class_label_prefix = []
        else:
            raise NotImplementedError

        super().__init__(
            name=f"Credit {self.split.value}",
            num_samples=30000,
            filename_or_path="UCI_Credit_Card.csv",
            features=disc_features + continuous_features,
            cont_features=continuous_features,
            s_prefix=s_prefix,
            sens_attr_spec=sens_attr_spec,
            class_label_prefix=class_label_prefix,
            class_label_spec=class_label_spec,
            discrete_only=self.discrete_only,
            discrete_feature_groups=disc_feature_groups,
        )



"""Script to generate the toy dataset."""
from pathlib import Path

import numpy as np
import pandas as pd


def main(seed: int, samples: int) -> None:
    """Make a toy dataset that can be used to check an algorithm runs."""
    rng = np.random.default_rng(seed)

    feat1_g1: np.ndarray = rng.standard_normal(size=samples // 2) - 0.1
    feat1_g2: np.ndarray = rng.standard_normal(size=samples // 2) + 0.5

    feat2_g1: np.ndarray = rng.standard_normal(samples // 2) * 0.9
    feat2_g2: np.ndarray = rng.standard_normal(samples // 2) * 1.1

    disc_1 = rng.choice(a=['a', 'b', 'c', 'd', 'e'], size=samples, replace=True)
    disc_2 = rng.choice(a=['x', 'y', 'z'], size=samples, replace=True)

    sensitive_attr_g1 = np.full(feat1_g1.shape, 0)
    sensitive_attr_g2 = np.full(feat1_g2.shape, 1)
    feat1 = np.concatenate((feat1_g1, feat1_g2))
    feat2 = np.concatenate((feat2_g1, feat2_g2))
    sensitive_attr = np.concatenate((sensitive_attr_g1, sensitive_attr_g2))

    # There has to be a better way of dong this.
    decision = rng.multinomial(
        n=1,
        pvals=rng.dirichlet(
            alpha=(
                abs(feat1).mean(),
                abs(feat2).mean(),
                abs(feat1).mean() + abs(feat2).mean(),
                abs(feat1_g1).mean(),
                abs(feat2_g2).mean(),
            ),
            size=samples,
        ).mean(axis=0),
        size=samples,
    ).argmax(axis=1)

    df = pd.DataFrame(
        data={
            "decision": decision,
            "sensitive-attr": sensitive_attr,
            "a1": feat1,
            "a2": feat2,
            "disc_1": disc_1,
            "disc_2": disc_2,
        }
    )
    cols_to_dum = ["sensitive-attr", "disc_1", "disc_2"]
    df = pd.get_dummies(df, columns=cols_to_dum)

    # Shuffle the data,
    df = df.sample(frac=1.0, random_state=seed).reset_index(drop=True)

    # Save the CSV
    df.to_csv(str(Path(__file__).parent / "nbt.csv"), index=False)


if __name__ == "__main__":
    SEED = 888
    SAMPLES = 400
    main(SEED, SAMPLES)



"""Transparently show how the Hearitage Health dataset was modified from the raw download."""
# The Heritage Health dataset. It needs some (mild) preprocessing before we can plug and play.

import pandas as pd

df = pd.read_csv("raw/health.csv")

# Drop columns which contains NaNs
print(f"Dropping columns: {df.columns[df.isna().any()].tolist()}")
df = df.dropna(axis=1)

# Add binary class column
df["Charlson>0"] = df["CharlsonIndexI_max"] > 0
df["Charlson>0"] = df["Charlson>0"].astype(int)

# Drop rows with missing age and sex
df = df[(df["age_MISS"] == 0) & (df["sexMISS"] == 0)]

# Shuffle the data
df = df.sample(frac=1.0, random_state=888).reset_index(drop=True)

# Save the CSV
df.to_csv("./health.csv", index=False)



"""Code is taken from https://github.com/equialgo/fairness-in-ml.

Original implementation is modified to handle regression and multi-class
classification problems
"""
from typing import Tuple
from typing_extensions import Literal, Self

import numpy as np
import pandas as pd
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from ranzen import implements

from ethicml import DataTuple
from ethicml.implementations.pytorch_common import (
    DeepModel,
    DeepRegModel,
    LinearModel,
    make_dataset_and_loader,
)


class Adversary(nn.Module):
    """Adversarial model."""

    def __init__(self, n_sensitive: int, n_y: int, n_hidden: int = 32):
        super().__init__()
        self.network = nn.Sequential(
            nn.Linear(n_y, n_hidden),
            nn.ReLU(),
            nn.Linear(n_hidden, n_hidden),
            nn.ReLU(),
            nn.Linear(n_hidden, n_hidden),
            nn.ReLU(),
            nn.Linear(n_hidden, n_sensitive),
        )

    @implements(nn.Module)
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.sigmoid(self.network(x))


def pretrain_adversary(
    adv: nn.Module,
    clf: nn.Module,
    data_loader: torch.utils.data.DataLoader,
    optimizer: torch.optim.Optimizer,
    criterion: nn.Module,
    lambdas: torch.Tensor,
) -> nn.Module:
    """Pretrain adversary."""
    for x, z, y in data_loader:
        p_y = clf(x).detach()
        adv.zero_grad()
        if len(p_y.size()) == 1:
            p_y = p_y.unsqueeze(dim=1)
        y_ = F.one_hot(y.long())
        in_adv = torch.cat((p_y, y_), 1)
        p_z = adv(in_adv)
        loss = (criterion(p_z, z.unsqueeze(1)) * lambdas).mean()
        loss.backward()
        optimizer.step()
    return adv


def train_loop(
    clf: nn.Module,
    adv: nn.Module,
    data_loader: torch.utils.data.DataLoader,
    clf_criterion: nn.Module,
    adv_criterion: nn.Module,
    clf_optimizer: torch.optim.Optimizer,
    adv_optimizer: torch.optim.Optimizer,
    lambdas: torch.Tensor,
) -> Tuple[nn.Module, nn.Module]:
    """Train model."""
    # Train adversary
    for x, z, y in data_loader:
        p_y = clf(x)
        if len(p_y.size()) == 1:
            p_y = p_y.unsqueeze(dim=1)
        adv.zero_grad()
        y_ = F.one_hot(y.long())
        in_adv = torch.cat((p_y, y_), 1)
        p_z = adv(in_adv)
        loss_adv = (adv_criterion(p_z, z.unsqueeze(1)) * lambdas).mean()
        loss_adv.backward()
        adv_optimizer.step()

    # Train predictor on single batch
    for x, z, y in data_loader:
        p_y = clf(x)
        if len(p_y.size()) == 1:
            p_y = p_y.unsqueeze(dim=1)
        y_ = F.one_hot(y.long())
        in_adv = torch.cat((p_y, y_), 1)
        p_z = adv(in_adv)  # TODO: This is unused?
        clf.zero_grad()
        p_z = adv(in_adv)
        loss_adv = (adv_criterion(p_z, z.unsqueeze(1)) * lambdas).mean()  # TODO: This is unused?
        clf_loss = (1.0 - lambdas) * clf_criterion(p_y, y.squeeze().long()) - (
            adv_criterion(adv(in_adv), z.unsqueeze(1)) * lambdas
        ).mean()
        clf_loss.backward()
        clf_optimizer.step()
        break

    return clf, adv


def train_regressor(
    clf: nn.Module,
    adv: nn.Module,
    data_loader: torch.utils.data.DataLoader,
    clf_criterion: nn.Module,
    adv_criterion: nn.Module,
    clf_optimizer: torch.optim.Optimizer,
    adv_optimizer: torch.optim.Optimizer,
    lambdas: torch.Tensor,
) -> Tuple[nn.Module, nn.Module]:
    """Train regression model."""
    # Train adversary
    for x, z, y in data_loader:
        p_y = clf(x)
        if len(p_y.size()) == 1:
            p_y = p_y.unsqueeze(dim=1)
        adv.zero_grad()
        in_adv = torch.cat((p_y, y), 1)
        p_z = adv(in_adv)
        loss_adv = (adv_criterion(p_z, z) * lambdas).mean()
        loss_adv.backward()
        adv_optimizer.step()

    # Train predictor on single batch
    for x, z, y in data_loader:
        p_y = clf(x)
        if len(p_y.size()) == 1:
            p_y = p_y.unsqueeze(dim=1)
        in_adv = torch.cat((p_y, y), 1)
        p_z = adv(in_adv)  # TODO: This is unused?
        clf.zero_grad()
        p_z = adv(in_adv)
        loss_adv = (adv_criterion(p_z, z) * lambdas).mean()  # TODO: This is unused?
        clf_loss = (1.0 - lambdas) * clf_criterion(p_y.squeeze(), y.squeeze()) - (
            adv_criterion(adv(in_adv), z) * lambdas
        ).mean()
        clf_loss.backward()
        clf_optimizer.step()
        break

    return clf, adv


def pretrain_classifier(
    clf: nn.Module,
    data_loader: torch.utils.data.DataLoader,
    optimizer: torch.optim.Optimizer,
    criterion: nn.Module,
) -> nn.Module:
    """Pretrain classification model."""
    for x, _, y in data_loader:
        clf.zero_grad()
        p_y = clf(x)
        loss = criterion(p_y, y.squeeze().long())
        loss.backward()
        optimizer.step()
    return clf


def pretrain_regressor(
    clf: nn.Module,
    data_loader: torch.utils.data.DataLoader,
    optimizer: torch.optim.Optimizer,
    criterion: nn.Module,
) -> nn.Module:
    """Pretrain regression model."""
    for x, _, y in data_loader:
        clf.zero_grad()
        p_y = clf(x)
        loss = criterion(p_y.squeeze(), y.squeeze())
        loss.backward()
        optimizer.step()
    return clf


class AdvDebiasingClassLearner:
    """Adversarial Debiasing classifier."""

    def __init__(
        self,
        lr: float,
        n_clf_epochs: int,
        n_adv_epochs: int,
        n_epoch_combined: int,
        cost_pred: nn.Module,
        in_shape: int,
        batch_size: int,
        model_type: Literal["deep_model", "linear_model"],
        num_classes: int,
        lambda_vec: float,
    ):

        self.lr = lr
        self.batch_size = batch_size
        self.in_shape = in_shape
        self.num_classes = num_classes

        self.model_type = model_type
        if self.model_type == "deep_model":
            self.clf: nn.Module = DeepModel(in_shape=in_shape, out_shape=num_classes)
        elif self.model_type == "linear_model":
            self.clf = LinearModel(in_shape=in_shape, out_shape=num_classes)
        else:
            raise NotImplementedError

        self.clf_criterion = cost_pred
        self.clf_optimizer = optim.Adam(self.clf.parameters(), lr=self.lr)

        self.n_clf_epochs = n_clf_epochs

        self.lambdas = torch.Tensor([lambda_vec])

        self.adv: nn.Module = Adversary(n_sensitive=1, n_y=num_classes * 2)
        self.adv_criterion = nn.BCELoss(reduction='none')
        self.adv_optimizer = optim.Adam(self.adv.parameters(), lr=self.lr)

        self.n_adv_epochs = n_adv_epochs

        self.n_epoch_combined = n_epoch_combined

    def fit(self, train: DataTuple, seed: int) -> Self:  # type: ignore[valid-type]
        """Fit."""
        train_data, train_loader = make_dataset_and_loader(
            train, batch_size=self.batch_size, shuffle=True, seed=seed, drop_last=True
        )

        for _ in range(self.n_clf_epochs):
            self.clf = pretrain_classifier(
                self.clf, train_loader, self.clf_optimizer, self.clf_criterion
            )

        for _ in range(self.n_adv_epochs):
            pretrain_adversary(
                self.adv,
                self.clf,
                train_loader,
                self.adv_optimizer,
                self.adv_criterion,
                self.lambdas,
            )

        for _ in range(1, self.n_epoch_combined):
            self.clf, self.adv = train_loop(
                self.clf,
                self.adv,
                train_loader,
                self.clf_criterion,
                self.adv_criterion,
                self.clf_optimizer,
                self.adv_optimizer,
                self.lambdas,
            )
        return self

    @torch.no_grad()
    def predict(self, x: pd.DataFrame) -> np.ndarray:
        """Predict."""
        x = torch.from_numpy(x.to_numpy()).float()
        self.clf.eval()
        yhat = self.clf(x)
        sm = nn.Softmax(dim=1)
        yhat = sm(yhat)
        yhat = yhat.detach().numpy()

        return yhat


class AdvDebiasingRegLearner:
    """Adversarial Debiasing Learner."""

    def __init__(
        self,
        lr: float,
        n_clf_epochs: int,
        n_adv_epochs: int,
        n_epoch_combined: int,
        cost_pred: nn.Module,
        in_shape: int,
        batch_size: int,
        model_type: Literal["deep_model", "linear_model"],
        out_shape: int,
        lambda_vec: float,
    ):

        self.lr = lr
        self.batch_size = batch_size
        self.in_shape = in_shape
        self.out_shape = out_shape

        self.model_type = model_type
        if self.model_type == "deep_model":
            self.clf: nn.Module = DeepRegModel(in_shape=in_shape, out_shape=out_shape)
        elif self.model_type == "linear_model":
            self.clf = LinearModel(in_shape=in_shape, out_shape=out_shape)
        else:
            raise NotImplementedError

        self.clf_criterion = cost_pred
        self.clf_optimizer = optim.Adam(self.clf.parameters(), lr=self.lr)

        self.N_CLF_EPOCHS = n_clf_epochs

        self.lambdas = torch.Tensor([lambda_vec])

        self.adv: nn.Module = Adversary(n_sensitive=1, n_y=out_shape + 1)
        self.adv_criterion = nn.BCELoss(reduce=False)
        self.adv_optimizer = optim.Adam(self.adv.parameters(), lr=self.lr)

        self.n_adv_epochs = n_adv_epochs

        self.n_epoch_combined = n_epoch_combined

    def fit(self, train: DataTuple, seed: int) -> Self:  # type: ignore[valid-type]
        """Fit."""
        # The features are X[:,1:]

        _, train_loader = make_dataset_and_loader(
            train, batch_size=self.batch_size, shuffle=True, seed=seed, drop_last=True
        )

        for _ in range(self.N_CLF_EPOCHS):
            self.clf = pretrain_regressor(
                self.clf, train_loader, self.clf_optimizer, self.clf_criterion
            )

        for _ in range(self.n_adv_epochs):
            pretrain_adversary(
                self.adv,
                self.clf,
                train_loader,
                self.adv_optimizer,
                self.adv_criterion,
                self.lambdas,
            )

        for _ in range(1, self.n_epoch_combined):
            self.clf, self.adv = train_regressor(
                self.clf,
                self.adv,
                train_loader,
                self.clf_criterion,
                self.adv_criterion,
                self.clf_optimizer,
                self.adv_optimizer,
                self.lambdas,
            )
        return self

    @torch.no_grad()
    def predict(self, x: pd.DataFrame) -> torch.Tensor:
        """Predict."""
        x = torch.from_numpy(x.to_numpy()).float()
        self.clf.eval()
        yhat = self.clf(x).squeeze().detach().numpy()
        if self.out_shape == 1:
            out = yhat
        else:
            out = 0 * yhat
            out[:, 0] = np.min(yhat, axis=1)
            out[:, 1] = np.max(yhat, axis=1)
        return out



"""Transparently show how the UCI Adult dataset was generated from the raw download."""

import numpy as np
import pandas as pd


def run_generate_crime() -> None:
    """Generate the UCI Communities and Crime dataset from scratch."""
    # We use data from the UCI repository directly and show how the final csv is generated.
    # To use the .arff file from K&C 2012, first download the .arff file from
    # https://sites.google.com/site/faisalkamiran/ which is referenced in their 2012 paper
    # available here: https://core.ac.uk/download/pdf/81728147.pdf
    # then, uncomment the following code.
    #
    # from scipy.io import arff
    # import pandas as pd
    # data = arff.loadarff('./raw/communities.arff')
    # data = pd.DataFrame(data[0])
    # for col in data.columns:
    #     if data[col].dtype == np.object:
    #         data[col] = data[col].astype(int)
    # data = data.replace(r"^\s*\?+\s*$", np.nan, regex=True).dropna()
    # data = data.sample(frac=1.0, random_state=888).reset_index(drop=True)
    # data.to_csv("./crime.csv", index=False)

    # Load the data
    data = pd.read_csv("raw/communities.data")

    columns = [
        "state",
        "county",
        "community",
        "communityname",
        "fold",
        "population",
        "householdsize",
        "racepctblack",
        "racePctWhite",
        "racePctAsian",
        "racePctHisp",
        "agePct12t21",
        "agePct12t29",
        "agePct16t24",
        "agePct65up",
        "numbUrban",
        "pctUrban",
        "medIncome",
        "pctWWage",
        "pctWFarmSelf",
        "pctWInvInc",
        "pctWSocSec",
        "pctWPubAsst",
        "pctWRetire",
        "medFamInc",
        "perCapInc",
        "whitePerCap",
        "blackPerCap",
        "indianPerCap",
        "AsianPerCap",
        "OtherPerCap",
        "HispPerCap",
        "NumUnderPov",
        "PctPopUnderPov",
        "PctLess9thGrade",
        "PctNotHSGrad",
        "PctBSorMore",
        "PctUnemployed",
        "PctEmploy",
        "PctEmplManu",
        "PctEmplProfServ",
        "PctOccupManu",
        "PctOccupMgmtProf",
        "MalePctDivorce",
        "MalePctNevMarr",
        "FemalePctDiv",
        "TotalPctDiv",
        "PersPerFam",
        "PctFam2Par",
        "PctKids2Par",
        "PctYoungKids2Par",
        "PctTeen2Par",
        "PctWorkMomYoungKids",
        "PctWorkMom",
        "NumIlleg",
        "PctIlleg",
        "NumImmig",
        "PctImmigRecent",
        "PctImmigRec5",
        "PctImmigRec8",
        "PctImmigRec10",
        "PctRecentImmig",
        "PctRecImmig5",
        "PctRecImmig8",
        "PctRecImmig10",
        "PctSpeakEnglOnly",
        "PctNotSpeakEnglWell",
        "PctLargHouseFam",
        "PctLargHouseOccup",
        "PersPerOccupHous",
        "PersPerOwnOccHous",
        "PersPerRentOccHous",
        "PctPersOwnOccup",
        "PctPersDenseHous",
        "PctHousLess3BR",
        "MedNumBR",
        "HousVacant",
        "PctHousOccup",
        "PctHousOwnOcc",
        "PctVacantBoarded",
        "PctVacMore6Mos",
        "MedYrHousBuilt",
        "PctHousNoPhone",
        "PctWOFullPlumb",
        "OwnOccLowQuart",
        "OwnOccMedVal",
        "OwnOccHiQuart",
        "RentLowQ",
        "RentMedian",
        "RentHighQ",
        "MedRent",
        "MedRentPctHousInc",
        "MedOwnCostPctInc",
        "MedOwnCostPctIncNoMtg",
        "NumInShelters",
        "NumStreet",
        "PctForeignBorn",
        "PctBornSameState",
        "PctSameHouse85",
        "PctSameCity85",
        "PctSameState85",
        "LemasSwornFT",
        "LemasSwFTPerPop",
        "LemasSwFTFieldOps",
        "LemasSwFTFieldPerPop",
        "LemasTotalReq",
        "LemasTotReqPerPop",
        "PolicReqPerOffic",
        "PolicPerPop",
        "RacialMatchCommPol",
        "PctPolicWhite",
        "PctPolicBlack",
        "PctPolicHisp",
        "PctPolicAsian",
        "PctPolicMinor",
        "OfficAssgnDrugUnits",
        "NumKindsDrugsSeiz",
        "PolicAveOTWorked",
        "LandArea",
        "PopDens",
        "PctUsePubTrans",
        "PolicCars",
        "PolicOperBudg",
        "LemasPctPolicOnPatr",
        "LemasGangUnitDeploy",
        "LemasPctOfficDrugUn",
        "PolicBudgPerPop",
        "ViolentCrimesPerPop",
    ]

    data.columns = pd.Index(columns)

    for col in data.columns:
        if data[col].dtype == np.object:  # type: ignore[attr-defined]
            data[col] = data[col].str.strip()

    # Drop NaNs
    data = data.replace(r"^\s*\?+\s*$", np.nan, regex=True).dropna(axis=1)

    # OHE
    data = pd.get_dummies(data, columns=["state"])

    # Add a new column, ">0.06black" as in:
    # Faisal Kamiran and Toon Calders. 2012. Data preprocessing techniques for classification without discrimination.
    # Knowledge and Information Systems 33, 1 (2012), 1–33.
    data[">0.06black"] = data["racepctblack"] > 0.06
    data[">0.06black"] = data[">0.06black"].astype(int)

    # Add a 'class' label of high violent crime as in K&C 2012
    data["high_crime"] = data["ViolentCrimesPerPop"] > 0.25
    data["high_crime"] = data["high_crime"].astype(int)

    # Shuffle the data
    data = data.sample(frac=1.0, random_state=888).reset_index(drop=True)

    # Save the CSV
    data.to_csv("./crime.csv", index=False)

    print(list(data.columns))


if __name__ == "__main__":
    run_generate_crime()



"""Generate biased subsets."""
from typing import Dict, Sequence, Tuple

import numpy as np
import pandas as pd
import teext as tx

from ethicml.utility.data_structures import DataTuple, concat

from .domain_adaptation import query_dt
from .splits import DataSplitter, ProportionalSplit

__all__ = [
    "BiasedDebiasedSubsets",
    "BiasedSubset",
    "get_biased_and_debiased_subsets",
    "get_biased_subset",
]


class BiasedSubset(DataSplitter):
    """Split the given data into a biased subset and a normal subset.

    :param mixing_factors: List of mixing factors; they are chosen based on the split ID
    :param unbiased_pcnt: how much of the data should be reserved for the unbiased subset
    :param seed: random seed for the splitting
    :param data_efficient: if True, try to keep as many data points as possible
    """

    def __init__(
        self,
        unbiased_pcnt: float,
        mixing_factors: Sequence[float] = (0,),
        seed: int = 42,
        data_efficient: bool = True,
    ):
        super().__init__()
        self.unbiased_pcnt = unbiased_pcnt
        self.mixing_factors = mixing_factors
        self.seed = seed
        self.data_efficient = data_efficient

    def __call__(
        self, data: DataTuple, split_id: int = 0
    ) -> Tuple[DataTuple, DataTuple, Dict[str, float]]:
        """Do BiasedSplit."""
        mixing_factor = self.mixing_factors[split_id]
        biased, unbiased = get_biased_subset(
            data, mixing_factor, self.unbiased_pcnt, self.seed, self.data_efficient
        )
        return biased, unbiased, {"mix_fact": mixing_factor}


def get_biased_subset(
    data: DataTuple,
    mixing_factor: float,
    unbiased_pcnt: float,
    seed: int = 42,
    data_efficient: bool = True,
) -> Tuple[DataTuple, DataTuple]:
    """Split the given data into a biased subset and a normal subset.

    The two subsets don't generally sum up to the whole set.

    Example behavior:

    * mixing_factor=0.0: in biased, s=y everywhere; unbiased is just a subset of `data`
    * mixing_factor=0.5: biased and unbiased are both just subsets of `data`
    * mixing_factor=1.0: in biased, s!=y everywhere; unbiased is just a subset of `data`

    :param data: data in form of a DataTuple
    :param mixing_factor: How much of the debiased data should be mixed into the biased subset? If this
                       factor is 0, the biased subset is maximally biased.
    :param unbiased_pcnt: how much of the data should be reserved for the unbiased subset
    :param seed: random seed for the splitting (Default: 42)
    :param data_efficient: if True, try to keep as many data points as possible (Default: True)
    :returns: biased and unbiased dataset
    """
    assert tx.is_percentage(mixing_factor), f"mixing_factor: {mixing_factor}"
    assert tx.is_percentage(unbiased_pcnt), f"unbiased_pcnt: {unbiased_pcnt}"
    s_name = data.s.name
    y_name = data.y.name
    assert isinstance(s_name, str) and isinstance(y_name, str)

    normal_subset, for_biased_subset = _random_split(data, first_pcnt=unbiased_pcnt, seed=seed)

    sy_equal, sy_opposite = _get_sy_equal_and_opp(for_biased_subset, s_name, y_name)

    mix_fact = mixing_factor  # how much of sy_opp should be mixed into the biased subset

    if data_efficient:
        if mix_fact < 0.5:
            sy_equal_fraction = 1.0
            sy_opp_fraction = 2 * mix_fact
        else:
            sy_equal_fraction = 2 * (1 - mix_fact)
            sy_opp_fraction = 1.0
    else:
        sy_equal_fraction = 1 - mix_fact
        sy_opp_fraction = mix_fact

    sy_equal_for_biased_ss, _ = _random_split(sy_equal, first_pcnt=sy_equal_fraction, seed=seed)
    sy_opp_for_biased_ss, _ = _random_split(sy_opposite, first_pcnt=sy_opp_fraction, seed=seed)

    biased_subset = concat([sy_equal_for_biased_ss, sy_opp_for_biased_ss], ignore_index=True)

    if mix_fact == 0.0:
        # s and y should be very correlated in the biased subset
        assert all(biased_subset.s == biased_subset.y)

    biased_subset = biased_subset.rename(f"{data.name} - Biased (tm={mixing_factor})")
    normal_subset = normal_subset.rename(f"{data.name} - Subset (tm={mixing_factor})")
    return biased_subset, normal_subset


class BiasedDebiasedSubsets(DataSplitter):
    """Split the given data into a biased subset and a debiased subset.

    :param mixing_factors: List of mixing factors; they are chosen based on the split ID
    :param unbiased_pcnt: how much of the data should be reserved for the unbiased subset
    :param seed: random seed for the splitting
    :param fixed_unbiased: if True, then the unbiased dataset is independent from the mixing factor
    """

    def __init__(
        self,
        unbiased_pcnt: float,
        mixing_factors: Sequence[float] = (0,),
        seed: int = 42,
        fixed_unbiased: bool = True,
    ):
        super().__init__()
        self.unbiased_pcnt = unbiased_pcnt
        self.mixing_factors = mixing_factors
        self.seed = seed
        self.fixed_unbiased = fixed_unbiased

    def __call__(
        self, data: DataTuple, split_id: int = 0
    ) -> Tuple[DataTuple, DataTuple, Dict[str, float]]:
        """Do Biased and Debiased Split."""
        mixing_factor = self.mixing_factors[split_id]
        biased, unbiased = get_biased_and_debiased_subsets(
            data, mixing_factor, self.unbiased_pcnt, self.seed, self.fixed_unbiased
        )
        return biased, unbiased, {"mix_fact": mixing_factor}


def get_biased_and_debiased_subsets(
    data: DataTuple,
    mixing_factor: float,
    unbiased_pcnt: float,
    seed: int = 42,
    fixed_unbiased: bool = True,
) -> Tuple[DataTuple, DataTuple]:
    """Split the given data into a biased subset and a debiased subset.

    In contrast to :func:`get_biased_subset()`, this function makes the unbiased subset *really*
    unbiased.

    The two subsets don't generally sum up to the whole set.

    Example behavior:

    * mixing_factor=0.0: in biased, s=y everywhere; in debiased, 50% s=y and 50% s!=y
    * mixing_factor=0.5: biased is just a subset of `data`; in debiased, 50% s=y and 50% s!=y
    * mixing_factor=1.0: in biased, s!=y everywhere; in debiased, 50% s=y and 50% s!=y

    :param data: data in form of a DataTuple
    :param mixing_factor: How much of the debiased data should be mixed into the biased subset? If this
                       factor is 0, the biased subset is maximally biased.
    :param unbiased_pcnt: how much of the data should be reserved for the unbiased subset
    :param seed: random seed for the splitting (Default: 42)
    :param fixed_unbiased: if True, then the unbiased dataset is independent from the mixing factor (Default: True)
    :returns: biased and unbiased dataset
    """
    assert tx.is_percentage(mixing_factor), f"mixing_factor: {mixing_factor}"
    assert tx.is_percentage(unbiased_pcnt), f"unbiased_pcnt: {unbiased_pcnt}"
    s_name = data.s.name
    y_name = data.y.name
    assert isinstance(s_name, str) and isinstance(y_name, str)
    sy_equal, sy_opposite = _get_sy_equal_and_opp(data, s_name, y_name)

    # how much of sy_equal should be reserved for the biased subset:

    if fixed_unbiased:
        sy_equal_for_debiased_ss, sy_equal_remaining = _random_split(
            sy_equal, first_pcnt=unbiased_pcnt, seed=seed
        )
        sy_opp_for_debiased_ss, sy_opp_remaining = _random_split(
            sy_opposite, first_pcnt=unbiased_pcnt, seed=seed
        )
        if mixing_factor < 0.5:
            sy_equal_fraction = 1.0
            sy_opp_fraction = 2 * mixing_factor
        else:
            sy_equal_fraction = 2 * (1 - mixing_factor)
            sy_opp_fraction = 1.0

        sy_equal_for_biased_ss, _ = _random_split(
            sy_equal_remaining, first_pcnt=sy_equal_fraction, seed=seed
        )
        sy_opp_for_biased_ss, _ = _random_split(
            sy_opp_remaining, first_pcnt=sy_opp_fraction, seed=seed
        )
    else:
        biased_pcnt = 1 - unbiased_pcnt
        sy_equal_for_biased_ss, sy_equal_for_debiased_ss = _random_split(
            sy_equal, first_pcnt=biased_pcnt * (1 - mixing_factor), seed=seed
        )
        sy_opp_for_biased_ss, sy_opp_for_debiased_ss = _random_split(
            sy_opposite, first_pcnt=biased_pcnt * mixing_factor, seed=seed
        )

    biased_subset = concat([sy_equal_for_biased_ss, sy_opp_for_biased_ss], ignore_index=True)

    # the debiased set is constructed from two sets of the same size
    min_size = min(len(sy_equal_for_debiased_ss), len(sy_opp_for_debiased_ss))

    def _get_equal_sized_subset(df: pd.DataFrame) -> pd.DataFrame:
        return df.sample(n=min_size, random_state=seed)  # type: ignore[return-value]

    debiased_subset_part1 = sy_equal_for_debiased_ss.apply_to_joined_df(_get_equal_sized_subset)
    debiased_subset_part2 = sy_opp_for_debiased_ss.apply_to_joined_df(_get_equal_sized_subset)
    debiased_subset = concat([debiased_subset_part1, debiased_subset_part2], ignore_index=True)

    # s and y should not be correlated in the debiased subset
    assert abs(debiased_subset.s.corr(debiased_subset.y)) < 0.5

    if mixing_factor == 0.0:
        # s and y should be very correlated in the biased subset
        assert biased_subset.s.corr(biased_subset.y) > 0.99

    biased_subset = biased_subset.rename(f"{data.name} - Biased (tm={mixing_factor})")
    debiased_subset = debiased_subset.rename(f"{data.name} - Debiased (tm={mixing_factor})")
    return biased_subset, debiased_subset


def _random_split(data: DataTuple, first_pcnt: float, seed: int) -> Tuple[DataTuple, DataTuple]:
    if len(data) == 0:
        return data, data
    splitter = ProportionalSplit(train_percentage=first_pcnt, start_seed=seed)
    return splitter(data)[:2]


def _get_sy_equal_and_opp(data: DataTuple, s_name: str, y_name: str) -> Tuple[DataTuple, DataTuple]:
    """Get the subset where s and y are equal and the subset where they are opposite.

    :param data:
    :param s_name:
    :param y_name:
    """
    s_values = np.unique(data.s.to_numpy())
    y_values = np.unique(data.y.to_numpy())
    assert len(s_values) == 2, "function only works with binary sensitive attribute"
    assert len(y_values) == 2, "function only works with binary labels"
    # we implicitly assume that the minimum value of s and y are equal; same for the maximum value
    s0, s1 = s_values
    y0, y1 = y_values
    # datapoints where s and y are the same
    s_and_y_equal = query_dt(
        data,
        f"(`{s_name}` == {s0} & `{y_name}` == {y0}) | (`{s_name}` == {s1} & `{y_name}` == {y1})",
    )
    # datapoints where they are not the same
    s_and_y_opposite = query_dt(
        data,
        f"(`{s_name}` == {s0} & `{y_name}` == {y1}) | (`{s_name}` == {s1} & `{y_name}` == {y0})",
    )
    return s_and_y_equal, s_and_y_opposite
